from rdflib import Graph, Literal, URIRef

from plastron.models.authorities import Agent
from plastron.models.umd import Item
from plastron.namespaces import dcterms, umdtype, rdfs
from plastron.rdfmapping.embed import embedded


def test_identifier_distinct_from_accession_number():
    rdf = (
        '@prefix dcterms: <http://purl.org/dc/terms/> .'
        '@prefix umdtype: <http://vocab.lib.umd.edu/datatype#> .'
        '<> dcterms:identifier "foo" .'
        '<> dcterms:identifier "1212XN1"^^umdtype:accessionNumber .'
    )
    base_uri = 'http://example.com/xyz'
    item = Item(
        graph=Graph().parse(data=rdf, format='turtle', publicID=base_uri),
        uri=base_uri
    )
    assert len(item.identifier) == 1
    assert item.identifier.value == Literal('foo')

    assert len(item.accession_number) == 1
    assert item.accession_number.value == Literal('1212XN1', datatype=umdtype.accessionNumber)

    assert (URIRef(base_uri), dcterms.identifier, Literal('foo')) in item.graph
    assert (URIRef(base_uri), dcterms.identifier, Literal('1212XN1', datatype=umdtype.accessionNumber)) in item.graph


def test_item_valid_with_only_required_fields():
    item = Item()

    # Only provide required fields
    item.identifier = 'test_item'
    item.object_type = 'http://purl.org/dc/dcmitype/Text'
    item.rights = 'http://vocab.lib.umd.edu/rightsStatement#InC-EDU'
    item.title = 'Test Item'
    assert item.is_valid


def test_audience_property():
    item = Item(uri=URIRef('http://example.com/foo'), audience=embedded(Agent)(label=Literal('John Doe')))

    audience_triples = list(item.graph.triples((URIRef('http://example.com/foo'), dcterms.audience, None)))
    assert len(audience_triples) == 1
    embedded_subject = audience_triples[0][2]
    label_triples = list(item.graph.triples((embedded_subject, rdfs.label, Literal('John Doe'))))
    assert len(label_triples) == 1



import logging
from collections.abc import Mapping
from os.path import abspath, dirname
from pathlib import Path
from typing import ItemsView
from urllib.error import HTTPError, URLError

from rdflib import Graph
from rdflib.term import URIRef, Literal

from plastron.namespaces import rdfs, dcterms, rdf, owl
from plastron.rdfmapping.descriptors import DataProperty, ObjectProperty
from plastron.rdfmapping.graph import TrackChangesGraph
from plastron.rdfmapping.resources import RDFResource

logger = logging.getLogger(__name__)

# Enable VOCABULARIES_DIR and VOCABULARIES to be overridden for tests
if 'VOCABULARIES_DIR' not in globals():
    VOCABULARIES_DIR = Path(dirname(abspath(__file__)))

if 'VOCABULARIES' not in globals():
    VOCABULARIES = {
        URIRef('http://purl.org/dc/dcmitype/'): 'dcmitype.ttl',
    }


class Vocabulary(Mapping):
    """Class representing an RDF vocabulary. Implements the `Mapping` abstract
    base class, so you can do all the following:

    ```pycon
    >>> vocab = Vocabulary('http://purl.org/dc/dcmitype/')

    >>> str(vocab)
    ''http://purl.org/dc/dcmitype/''

    >>> len(vocab)
    12

    >>> for term_uri in vocab:
    ...     print(term_uri)
    http://purl.org/dc/dcmitype/StillImage
    http://purl.org/dc/dcmitype/MovingImage
    http://purl.org/dc/dcmitype/Service
    http://purl.org/dc/dcmitype/Sound
    http://purl.org/dc/dcmitype/PhysicalObject
    http://purl.org/dc/dcmitype/Software
    http://purl.org/dc/dcmitype/Text
    http://purl.org/dc/dcmitype/Dataset
    http://purl.org/dc/dcmitype/Collection
    http://purl.org/dc/dcmitype/Event
    http://purl.org/dc/dcmitype/Image
    http://purl.org/dc/dcmitype/InteractiveResourcee

    >>> 'Image' in vocab
    True

    >>> URIRef('http://purl.org/dc/dcmitype/Image') in vocab
    True
    ```

    Membership checks and item lookups can both either use the full `URIRef` object
    representing a term, or just the local string portion of the URI that comes after
    the base namespace URI.

    Item lookup takes a subject URI and returns a dictionary where the keys are predicate
    URIs and the values are the objects (Literals or URIs) for statements with the given
    subject URI.

    By making use of `plastron.namespaces`, it is easy to retrieve details from
    vocabulary terms:

    ```pycon
    >>> from plastron.namespaces import rdfs

    >>> vocab['Image'][rdfs.label]
    rdflib.term.Literal('Image', lang='en')

    >>> str(vocab['Image'][rdfs.label])
    'Image'

    >>> for term_uri, term in vocab.items():
    ...     print(term[rdfs.label])
    Still Image
    Moving Image
    Service
    Sound
    Physical Object
    Software
    Text
    Dataset
    Collection
    Event
    Image
    Interactive Resource
    ```

    `Vocabulary` objects can also be used in the `values_from` attribute of RDF
    property mapping fields.
    """
    def __init__(self, uri: URIRef | str):
        self.uri = URIRef(uri)

    @property
    def term_uris(self):
        """Returns the set of URIs for all the terms in this vocabulary."""
        return set(get_vocabulary_graph(self.uri).subjects()) - {self.uri}

    def _term_uri(self, item) -> URIRef:
        return item if isinstance(item, URIRef) else URIRef(self.uri + item)

    def term_graph(self, term: URIRef | str) -> TrackChangesGraph:
        """Returns a graph containing all the triples with the given `term`
        as their subject."""
        graph = TrackChangesGraph()
        for triple in get_vocabulary_graph(self.uri).triples((self._term_uri(term), None, None)):
            graph.add(triple)
        return graph

    def __str__(self):
        return str(self.uri)

    def __len__(self) -> int:
        return len(self.term_uris)

    def __contains__(self, item) -> bool:
        return self._term_uri(item) in self.term_uris

    def __iter__(self):
        yield from self.term_uris

    def __getitem__(self, item):
        term_uri = self._term_uri(item)
        if term_uri not in self:
            raise KeyError(item)
        return {p: o for p, o in get_vocabulary_graph(self.uri).predicate_objects(term_uri)}

    def items(self) -> ItemsView[URIRef, dict]:
        """Returns an `ItemsView` mapping each term URI to a dictionary mapping
        of predicates to objects. Allows for easy iteration over the vocabulary
        as if it were a dictionary."""
        return {term_uri: self[term_uri] for term_uri in self}.items()

    def find(self, p: URIRef, o: URIRef | Literal) -> dict:
        """Finds the first term with a triple matching the given predicate `p`
        and object `o` in the vocabulary. Raises a `KeyError` if no such term
        can be found."""
        for _, term in self.items():
            if o in term[p]:
                return term
        else:
            raise KeyError(f'{p} {o}')


class VocabularyTerm(RDFResource):
    vocab: Vocabulary = None
    label = DataProperty(rdfs.label, required=True)
    description = DataProperty(dcterms.description)
    value = DataProperty(rdf.value)
    comment = DataProperty(rdfs.comment)
    same_as = ObjectProperty(owl.sameAs, repeatable=True)

    @classmethod
    def from_vocab(cls, vocab: Vocabulary, name: str = None) -> type:
        """Create a new subclass of `VocabularyTerm` with the `vocab` property
        set to the given `Vocabulary`. If `name` is not given, a name for the
        new class is generated by taking the last segment of the vocabulary
        URI and ensuring it starts with an uppercase letter.

        ```pycon
        >>> from plastron.validation.vocabularies import Vocabulary, VocabularyTerm

        >>> formats = Vocabulary('http://vocab.lib.umd.edu/form#')

        >>> new_cls = VocabularyTerm.from_vocab(formats)

        >>> new_cls.__name__
        'FormVocabularyTerm'

        >>> new_cls_with_name = VocabularyTerm.from_vocab(formats, 'UMDFormatsTerm')

        >>> new_cls_with_name.__name__
        'UMDFormatsTerm'
        ```
        """
        if name is None:
            vocab_name = vocab.uri.rstrip('/#').rsplit('/', 1)[1]
            name = vocab_name[0].upper() + vocab_name[1:] + 'VocabularyTerm'

        return type(name, (VocabularyTerm,), {'vocab': vocab})

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        if self.vocab is not None:
            # we just want the part with the term as the subject
            self._graph = self.vocab.term_graph(self.uri)


def get_vocabulary_graph(vocab_uri: URIRef) -> Graph:
    graph = Graph()
    # first check locally available vocabularies
    if vocab_uri in VOCABULARIES:
        try:
            graph.parse(
                location=str(VOCABULARIES_DIR / VOCABULARIES[vocab_uri]),
                format='turtle'
            )
            return graph
        except FileNotFoundError as e:
            logger.warning(f'Local version of {vocab_uri} not found: {e}')
            logger.info('Falling back to remote retrieval')

    # otherwise, fall back to remote retrieval
    try:
        graph.parse(location=vocab_uri)
    except (URLError, HTTPError) as e:
        raise RuntimeError(f'Unable to retrieve vocabulary from {vocab_uri}: {e}') from e

    return graph


class ControlledVocabularyProperty(ObjectProperty):
    """Specialized subclass of `plastron.rdfmapping.descriptors.ObjectProperty`
    for object properties whose values should come from a single controlled
    vocabulary. The following two property declarations are equivalent:

    ```python
    object_type = ControlledVocabularyProperty(
        dcterms.type,
        required=True,
        vocab=DCMI_TYPES,
    )

    object_type = ObjectProperty(
        dcterms.type,
        required=True,
        values_from=DCMI_TYPES,
        cls=VocabularyTerm.from_vocab(DCMI_TYPES),
    )
    ```
    """
    def __init__(self, predicate: URIRef, vocab: Vocabulary, **kwargs):
        super().__init__(
            predicate,
            **kwargs,
            values_from=vocab,
            cls=VocabularyTerm.from_vocab(vocab),
        )



from plastron.namespaces import rdfs, owl
from plastron.rdfmapping.descriptors import DataProperty, ObjectProperty
from plastron.rdfmapping.resources import RDFResource
from plastron.validation.vocabularies import Vocabulary

DCMI_TYPES = Vocabulary('http://purl.org/dc/dcmitype/')
UMD_RIGHTS_STATEMENTS = Vocabulary('http://vocab.lib.umd.edu/rightsStatement#')
UMD_FORMATS = Vocabulary('http://vocab.lib.umd.edu/form#')
UMD_ARCHIVAL_COLLECTIONS = Vocabulary('http://vocab.lib.umd.edu/collection#')
UMD_PRESENTATION_SETS = Vocabulary('http://vocab.lib.umd.edu/set#')
UMD_TERMS_OF_USE_STATEMENTS = Vocabulary('http://vocab.lib.umd.edu/termsOfUse#')


class LabeledThing(RDFResource):
    label = DataProperty(rdfs.label, required=True)
    same_as = ObjectProperty(owl.sameAs)


class Agent(LabeledThing):
    pass


class Subject(LabeledThing):
    pass


class Place(LabeledThing):
    pass



import logging
from datetime import datetime
from uuid import uuid4

from rdflib import Graph, URIRef

from plastron.client import ResourceURI, Client, ClientError
from plastron.rdf import rdf


class Resource(rdf.Resource):
    """Class representing a Linked Data Platform Resource (LDPR)
    A HTTP resource whose state is represented in any way that conforms to the
    simple lifecycle patterns and conventions in section 4. Linked Data Platform
    Resources."""

    @classmethod
    def from_repository(cls, client: Client, uri, include_server_managed=True):
        _, graph = client.get_graph(uri, include_server_managed=include_server_managed)
        obj = cls.from_graph(graph, subject=uri)
        obj.uri = uri
        obj.path = obj.uri[len(client.endpoint.url):]

        # get the description URI
        obj.resource = ResourceURI(uri, client.get_description_uri(uri))

        # mark as created and updated so that the create_object and update_object
        # methods doesn't try to modify it
        obj.created = True
        obj.updated = True

        return obj

    def __init__(self, uri='', **kwargs):
        super().__init__(uri=uri, **kwargs)
        self.annotations = []
        self.extra = Graph()
        self.created = False
        self.updated = False
        self.path = None
        self.container_path = None
        self.resource = None
        self.uuid = None
        self.creation_timestamp = None
        self.logger = logging.getLogger(
            __name__ + '.' + self.__class__.__name__
        )

    def __str__(self):
        if hasattr(self, 'title') and self.title is not None:
            return str(self.title)
        else:
            return repr(self)

    def load(self, repository: Client):
        return self.read(repository.get_graph(self.uri)[1])

    def create(self, client: Client, container_path=None, slug=None, headers=None, recursive=True, **kwargs):
        if not self.created and not self.exists_in_repo(client):

            self.logger.info(f"Creating {self}...")
            if headers is None:
                headers = {}
            if slug is not None:
                headers['Slug'] = slug
            try:
                self.resource = client.create(container_path=container_path, headers=headers, **kwargs)
                self.uri = URIRef(self.resource.uri)

                # if client.is_forwarded():
                #     # Reverse forwarded URL back to fcrepo URL
                #     parsed_resource_uri = urlsplit(self.resource.uri)
                #     parsed_path = parsed_resource_uri.path
                #     self.path = parsed_path[len(client.endpoint_base_path):]
                # else:
                self.path = self.resource.uri[len(client.endpoint.url):]

                self.created = True
                # TODO: get this from the response headers
                self.creation_timestamp = datetime.now()
                # TODO: get the fedora:parent
                self.container_path = container_path or client.endpoint.relpath
                self.logger.info(f"Created {self}")
                self.uuid = str(self.uri).rsplit('/', 1)[-1]
                self.logger.info(f'URI: {self.uri}')
                self.create_fragments()
            except ClientError as e:
                self.logger.error(f"Failed to create {self}: {e}")
                raise
        else:
            self.created = True
            self.logger.debug(f'Object "{self}" exists. Skipping.')

        if recursive:
            client.create_annotations(self)

    def create_fragments(self):
        for obj in self.embedded_objects():
            obj.uuid = uuid4()
            obj.uri = URIRef('{0}#{1}'.format(self.uri, obj.uuid))
            obj.created = True

    def patch(self, client, sparql_update):
        headers = {'Content-Type': 'application/sparql-update'}
        self.logger.info(f"Updating {self}")
        response = client.patch(self.resource.description_uri, data=sparql_update, headers=headers)
        if response.status_code == 204:
            self.logger.info(f"Updated {self}")
            self.updated = True
            return response
        else:
            self.logger.error(f"Failed to update {self}")
            self.logger.error(sparql_update)
            self.logger.error(response.text)
            raise ClientError(response)

    # update existing repo object with SPARQL update
    def update(self, client, recursive=True):
        if not self.updated:
            # Ensure item being updated has a ResourceURI
            if self.resource is None:
                self.resource = ResourceURI(self.uri, client.get_description_uri(self.uri))

            self.patch(client, client.build_sparql_update(insert_graph=self.graph()))
            if recursive:
                # recursively update an object and all its components and files
                for obj in self.linked_objects():
                    obj.update(client)
                # update any annotations
                for annotation in self.annotations:
                    annotation.update(client)

    # check for the existence of a local object in the repository
    def exists_in_repo(self, client: Client):
        if str(self.uri).startswith(client.endpoint.url):
            return client.exists(str(self.uri))
        else:
            return False

    # add arbitrary additional triples provided in a file
    def add_extra_properties(self, triples_file, rdf_format):
        self.extra.parse(
            source=triples_file, format=rdf_format, publicID=self.uri
        )

    # show the object's graph, serialized as turtle
    def print_graph(self):
        print(self.graph().serialize(format="turtle").decode())

    # called after creation of object in repo
    def post_creation_hook(self):
        pass

    def update_annotations(self, client: Client):
        for annotation in self.annotations:
            annotation.update(client)


class RdfSource(Resource):
    """Class representing a Linked Data Platform RDF Source (LDP-RS)
    An LDPR whose state is fully represented in RDF, corresponding to an RDF
    graph. See also the term RDF Source from [rdf11-concepts]."""
    pass


class NonRdfSource(Resource):
    """Class representing a Linked Data Platform Non-RDF Source (LDP-NR)
    An LDPR whose state is not represented in RDF. For example, these can be
    binary or text documents that do not have useful RDF representations."""
    @classmethod
    def from_source(cls, source=None, **kwargs):
        obj = cls(**kwargs)
        obj.source = source
        return obj


class Container(RdfSource):
    """Class representing a Linked Data Platform Container (LDPC)
    A LDP-RS representing a collection of linked documents (RDF Document
    [rdf11-concepts] or information resources [WEBARCH]) that responds to client
    requests for creation, modification, and/or enumeration of its linked
    members and documents, and that conforms to the simple lifecycle patterns
    and conventions in section 5. Linked Data Platform Containers."""
    pass


class BasicContainer(Container):
    """Class representing a Linked Data Platform Basic Container (LDP-BC)
    An LDPC that defines a simple link to its contained documents (information
    resources) [WEBARCH]."""
    pass


class DirectContainer(Container):
    """Class representing a Linked Data Platform Direct Container (LDP-DC)
    An LDPC that adds the concept of membership, allowing the flexibility of
    choosing what form its membership triples take, and allows members to be any
    resources [WEBARCH], not only documents."""
    pass


class IndirectContainer(Container):
    """Class representing a Linked Data Platform Indirect Container (LDP-IC)
    An LDPC similar to a LDP-DC that is also capable of having members whose
    URIs are based on the content of its contained documents rather than the
    URIs assigned to those documents."""
    pass



from http import HTTPStatus
import json
import httpretty
import pytest
from rdflib import Literal

from plastron.handles import HandleBearingResource, HandleServerError, HandleServiceClient, HandleInfo
from plastron.namespaces import umdtype


@pytest.fixture
def handle():
    return HandleInfo(exists=True, prefix='1903.1', suffix='123', url='http://example.com/foobar')


def test_handle_attributes(handle):
    assert handle.prefix == '1903.1'
    assert handle.suffix == '123'
    assert handle.url == 'http://example.com/foobar'
    assert handle.hdl_uri == 'hdl:1903.1/123'
    assert str(handle) == '1903.1/123'


def test_handle_bearing_resource():
    obj = HandleBearingResource(handle=Literal('hdl:1903.1/123', datatype=umdtype.handle))
    obj.apply_changes()
    assert obj.is_valid
    assert obj.handle.value == Literal('hdl:1903.1/123', datatype=umdtype.handle)


def test_invalid_handle_bearing_resource():
    obj = HandleBearingResource(handle=Literal('not-a-handle', datatype=umdtype.handle))
    assert not obj.is_valid
    assert obj.validate()['handle'].message == 'is not a handle URI'


@pytest.fixture
def handle_client():
    return HandleServiceClient('http://handle-local:3000', jwt_token='TOKEN')


@httpretty.activate
def test_get_handle_error(handle_client):
    httpretty.register_uri(
        httpretty.GET,
        uri='http://handle-local:3000/handles/exists',
        status=HTTPStatus.BAD_REQUEST,
    )
    with pytest.raises(HandleServerError):
        handle_client.find_handle('http://example.com/foobar')


@httpretty.activate
def test_get_handle_does_not_exist(handle_client):
    httpretty.register_uri(
        httpretty.GET,
        uri='http://handle-local:3000/handles/exists',
        body=json.dumps({'exists': False})
    )
    assert not handle_client.find_handle('http://example.com/foobar').exists


@httpretty.activate
def test_get_handle_exists(handle_client):
    httpretty.register_uri(
        httpretty.GET,
        uri='http://handle-local:3000/handles/exists',
        body=json.dumps({'exists': True, 'prefix': '1903.1', 'suffix': '123', 'url': 'http://example.com/foobar'})
    )
    handle = handle_client.find_handle('http://example.com/foobar')
    assert handle.prefix == '1903.1'
    assert handle.suffix == '123'
    assert handle.url == 'http://example.com/foobar'


@httpretty.activate
def test_create_handle_success(handle_client):
    httpretty.register_uri(
        httpretty.POST,
        uri='http://handle-local:3000/handles',
        body=json.dumps(
            {'suffix': '123', 'request': {'url': 'http://example.com/foobar', 'prefix': '1903.1'}}
        )
    )
    handle = handle_client.create_handle(repo_id='http://localhost/fcrepo/foobar', url='http://example.com/foobar')
    assert handle.prefix == '1903.1'
    assert handle.suffix == '123'
    assert handle.url == 'http://example.com/foobar'


@httpretty.activate
def test_create_handle_error(handle_client):
    httpretty.register_uri(
        httpretty.POST,
        uri='http://handle-local:3000/handles',
        status=HTTPStatus.BAD_REQUEST,
    )
    with pytest.raises(HandleServerError):
        handle_client.create_handle(repo_id='http://localhost/fcrepo/foobar', url='http://example.com/foobar')


@httpretty.activate
def test_update_handle(handle, handle_client):
    httpretty.register_uri(
        httpretty.PATCH,
        uri='http://handle-local:3000/handles/1903.1/123',
        body=json.dumps(
            {'suffix': '123', 'request': {'url': 'http://example.com/foobar', 'prefix': '1903.1'}}
        )
    )
    updated_handle = handle_client.update_handle(handle_info=handle, url='http://example.com/new-url')
    assert updated_handle.url == 'http://example.com/new-url'


@httpretty.activate
def test_update_handle_error(handle, handle_client):
    httpretty.register_uri(
        httpretty.PATCH,
        uri='http://handle-local:3000/handles/1903.1/123',
        status=HTTPStatus.BAD_REQUEST,
    )
    with pytest.raises(HandleServerError):
        handle_client.update_handle(handle_info=handle, url='http://example.com/new-url')



import sys

from rdflib import Graph, URIRef, Literal

import plastron.validation.rules
from plastron.validation import ValidationError
from plastron.namespaces import rdf

# alias the rdflib Namespace
ns = rdf


def rdf_class(*types):
    def add_types(cls):
        cls.rdf_types.update(types)
        return cls
    return add_types


# metaclass that makes copies of specific class attributes
# as we go down the inheritance hierarchy
class Meta(type):
    def __new__(mcs, name, bases, dct):
        cls = super().__new__(mcs, name, bases, dct)
        # name and URI to property type class lookups
        cls.name_to_prop = {}
        cls.uri_to_prop = {}
        # list of property type classes for this class
        cls.prop_types = []
        # RDF types
        cls.rdf_types = set()
        # copy values from all base classes
        for base in bases:
            cls.name_to_prop.update(base.name_to_prop)
            cls.uri_to_prop.update(base.uri_to_prop)
            cls.prop_types.extend(base.prop_types)
            cls.rdf_types.update(base.rdf_types)
        return cls


class RDFProperty:
    uri = None

    def __init__(self):
        self.values = []

    def append(self, other):
        if other not in self.values:
            self.values.append(other)

    def update(self, new_values):
        # take the set differences to find deleted and inserted values
        old_values_set = set(self.values)
        new_values_set = set(new_values)
        self.values = new_values
        deleted_values = old_values_set - new_values_set
        inserted_values = new_values_set - old_values_set
        # return the sets so the caller could construct a SPARQL update
        return deleted_values, inserted_values

    def __str__(self):
        return ' '.join(map(str, self.values))

    def __iter__(self):
        return iter(self.values)

    def __len__(self):
        return len(self.values)

    def __getattr__(self, item):
        if len(self.values) == 0:
            return None
        elif len(self.values) > 1:
            raise AttributeError(f'Multiple values for attribute {item} of {self}')
        else:
            return getattr(self.values[0], item)

    def __getitem__(self, item):
        return self.values[item]

    def triples(self, subject):
        for value in self.values:
            if value is not None:
                yield subject, self.uri, self.get_term(value)


class RDFDataProperty(RDFProperty):
    @classmethod
    def get_term(cls, value):
        return value if isinstance(value, Literal) else Literal(value, datatype=cls.datatype)


class RDFObjectProperty(RDFProperty):
    is_embedded = False
    obj_class = None

    def __getitem__(self, item):
        if isinstance(item, URIRef):
            for obj in self.values:
                if obj.uri == item:
                    return obj
            else:
                raise IndexError(f'Cannot find object by URI {item}')
        else:
            return super().__getitem__(item)

    def append_new(self, **kwargs):
        obj = self.obj_class(**kwargs)
        self.append(obj)

    @classmethod
    def get_term(cls, value):
        if hasattr(value, 'uri'):
            return URIRef(value.uri)
        elif isinstance(value, URIRef):
            return value
        elif isinstance(value, str):
            return URIRef(value)
        else:
            raise ValueError('Expecting a URIRef or an object with a uri attribute')


def data_property(name, uri, datatype=None):
    def add_property(cls):
        cls.add_data_property(name, uri, datatype)
        return cls
    return add_property


def object_property(name, uri, embed=False, obj_class=None):
    def add_property(cls):
        cls.add_object_property(name, uri, embed, obj_class)
        return cls
    return add_property


@object_property('rdf_type', rdf.type)
class Resource(metaclass=Meta):
    @classmethod
    def from_graph(cls, graph, subject=''):
        return cls(uri=subject).read(graph)

    @classmethod
    def add_data_property(cls, name, uri, datatype=None):
        type_name = f'{cls.__name__}.{name}'
        prop_type = type(type_name, (RDFDataProperty,), {
            'name': name,
            'uri': uri,
            'datatype': datatype
        })
        cls.name_to_prop[name] = prop_type
        cls.uri_to_prop[uri, datatype] = prop_type
        cls.prop_types.append(prop_type)

    @classmethod
    def add_object_property(cls, name, uri, embed=False, obj_class=None):
        type_name = f'{cls.__name__}.{name}'
        prop_type = type(type_name, (RDFObjectProperty,), {
            'name': name,
            'uri': uri,
            'is_embedded': embed,
            'obj_class': obj_class
        })
        cls.name_to_prop[name] = prop_type
        # object properties never have datatypes
        cls.uri_to_prop[uri, None] = prop_type
        cls.prop_types.append(prop_type)

    def __init__(self, uri=None, **kwargs):
        if uri is None:
            uri = ''
        self.uri = URIRef(uri)
        self.props = {}
        self.unmapped_triples = []

        for prop_type in self.prop_types:
            prop = prop_type()
            self.props[prop.name] = prop

        for rdf_type in self.rdf_types:
            self.rdf_type.append(rdf_type)

        for key, value in kwargs.items():
            setattr(self, key, value)

    def read(self, graph):
        for (s, p, o) in graph.triples((self.uri, None, None)):
            datatype = o.datatype if isinstance(o, Literal) else None
            if (p, datatype) in self.uri_to_prop:
                prop_type = self.uri_to_prop[p, datatype]
            elif (p, None) in self.uri_to_prop:
                # fall back to an untyped property
                prop_type = self.uri_to_prop[p, None]
            else:
                prop_type = None

            if prop_type is not None:
                if issubclass(prop_type, RDFObjectProperty) and prop_type.obj_class is not None:
                    obj = prop_type.obj_class(uri=o)
                    # recursively read embedded objects whose triples should be part of the same graph
                    if prop_type.is_embedded:
                        obj.read(graph)
                else:
                    obj = o
                getattr(self, prop_type.name).append(obj)
            else:
                self.unmapped_triples.append((s, p, o))
        return self

    def __getattr__(self, name):
        if name in self.name_to_prop:
            return self.props[name]
        else:
            raise AttributeError(f"No predicate mapped to {name}")

    def __setattr__(self, name, value):
        if name in self.name_to_prop:
            if not isinstance(value, list):
                value = [value]
            # TODO: wrap these values in rdflib classes?
            self.props[name].values = value
        else:
            # attribute names that aren't mapped to RDF properties
            # just get set like normal attributes
            super().__setattr__(name, value)

    def properties(self):
        return [prop for prop in self.props.values()]

    def data_properties(self):
        return [prop for prop in self.props.values() if isinstance(prop,
                                                                   RDFDataProperty)]

    def object_properties(self):
        return [prop for prop in self.props.values() if isinstance(prop,
                                                                   RDFObjectProperty)]

    def embedded_objects(self):
        for prop in [prop for prop in self.object_properties() if prop.is_embedded]:
            for v in prop.values:
                # recursively expand embedded objects
                if hasattr(v, 'embedded_objects'):
                    for vo in v.embedded_objects():
                        yield vo
                yield v

    def linked_objects(self):
        for prop in [prop for prop in self.object_properties() if not prop.is_embedded]:
            for v in [v for v in prop.values if hasattr(v, 'uri')]:
                yield v

    def graph(self, nsm=None):
        subject = URIRef(self.uri)
        graph = Graph(namespace_manager=nsm)

        for prop in self.properties():
            for (s, p, o) in prop.triples(self.uri):
                graph.add((s, p, o))

        # add graphs for all objects that should be embedded in the graph
        # for this resource; typically, these are resources that are have
        # hash URI identifiers
        for obj in self.embedded_objects():
            if obj is not None:
                graph = graph + obj.graph()

        # any triples that were loaded from the source graph but that aren't
        # mapped to specific Python attributes
        for (s, p, o) in self.unmapped_triples:
            graph.add((subject, p, o))

        return graph

    def print(self, format='turtle', file=sys.stdout, nsm=None):
        print(self.graph(nsm=nsm).serialize(format=format).decode(), file=file)

    def validate(self, ruleset=None):
        if ruleset is None:
            try:
                ruleset = self.VALIDATION_RULESET
            except AttributeError:
                raise ValidationError(f'No ruleset given, and "{self}" does not have a default ruleset')
        result = ResourceValidationResult(self)
        for field, rules in ruleset.items():
            for rule_name, arg in rules.items():
                rule = getattr(plastron.validation.rules, rule_name)
                prop = getattr(self, field)
                if isinstance(prop, RDFObjectProperty) and prop.is_embedded:
                    for obj in prop.values:
                        obj.validate(prop, result)
                else:
                    if rule(prop, arg):
                        result.passes(prop, rule, arg)
                    else:
                        result.fails(prop, rule, arg)
        return result


class ResourceValidationResult:
    def __init__(self, resource):
        self.resource = resource
        self.outcomes = []

    def __bool__(self):
        return self.is_valid()

    def is_valid(self):
        return len(list(self.failed())) == 0

    def passes(self, prop, rule, expected):
        self.outcomes.append((prop, 'passed', rule, expected))

    def fails(self, prop, rule, expected):
        self.outcomes.append((prop, 'failed', rule, expected))

    def passed(self):
        for prop, status, rule, expected in self.outcomes:
            if status != 'passed':
                continue
            yield prop.name, status, rule.__name__, getattr(expected, '__name__', expected)

    def failed(self):
        for prop, status, rule, expected in self.outcomes:
            if status != 'failed':
                continue
            yield prop.name, status, rule.__name__, getattr(expected, '__name__', expected)



import pytest as pytest
from rdflib import Literal, URIRef

from plastron.models.poster import Poster

base_uri = 'http://example.com/xyz'


def test_poster_invalid_with_no_fields():
    poster = Poster()
    assert not poster.is_valid


def test_poster_valid_with_only_required_fields():
    poster = Poster()

    # Only provide required fields
    poster.identifier = 'test_poster'
    poster.title = 'Test Poster'
    poster.language = 'en'
    poster.part_of = 'http://fedora.info/definitions/v4/repository#inaccessibleResource'
    poster.type = Literal('http://purl.org/dc/dcmitype/Image')
    poster.format = "Test Poster Format"
    poster.locator = 'NZK120'
    poster.rights = 'http://rightsstatements.org/vocab/NoC-US/1.0/'

    assert poster.is_valid


@pytest.mark.parametrize(
    ('subject_value', 'expected_validity'),
    [
        ([], True),
        ([Literal('single')], True),
        ([Literal('one'), Literal('two')], True),
    ]
)
def test_poster_subject_field_literals(subject_value, expected_validity):
    poster = Poster(subject=subject_value)
    assert bool(poster.subject.is_valid) == expected_validity


@pytest.mark.parametrize(
    'value',
    [
        ['foo'],
        [12345],
        [URIRef('http://example.com/foo')],
    ]
)
def test_poster_subject_field_non_literal_fails(value):
    with pytest.raises(TypeError):
        Poster(subject=value)



from PIL import Image, UnidentifiedImageError
from lxml import etree
from lxml.etree import XMLSyntaxError

from plastron.ocr.alto import ALTOResource
from plastron.ocr.core import OCRError, UnrecognizedOCRFormatError, OCRResource, ImageFileError, OCRFileError
from plastron.ocr.hocr import HOCRResource
from plastron.repo import BinaryResource, RepositoryError


class ImageWithOCR:
    def __init__(self, image_file: BinaryResource, ocr_file: BinaryResource):
        self.image_file = image_file
        self.ocr_file = ocr_file

    def get_ocr_resource(self) -> OCRResource:
        if self.ocr_file is None:
            raise OCRFileError('No OCR file specified')

        try:
            with self.ocr_file.open() as fh:
                doc = etree.parse(fh)
        except (RepositoryError, OSError, XMLSyntaxError) as e:
            raise OCRFileError(f'Cannot read OCR file {self.ocr_file.url}') from e

        if self.image_file is None:
            raise ImageFileError('No image file specified')

        try:
            with self.image_file.open() as fh:
                img = Image.open(fh)
                resolution = img.info['dpi']
        except (RepositoryError, FileNotFoundError, UnidentifiedImageError) as e:
            raise ImageFileError(f'Cannot read image file {self.image_file.url}') from e
        except KeyError:
            raise ImageFileError(f'Cannot read image resolution from {self.image_file.url}')

        root = doc.getroot()
        if root.tag == '{http://www.loc.gov/standards/alto/ns-v2#}alto':
            # ALTO XML
            return ALTOResource(doc, resolution)
        elif root.tag == '{http://www.w3.org/1999/xhtml}html':
            # hOCR
            return HOCRResource(doc, resolution)
        else:
            raise UnrecognizedOCRFormatError



from typing import Type, Dict

from importlib_metadata import entry_points
from rdflib import URIRef

from plastron.rdfmapping.resources import RDFResourceBase, RDFResource

PLUGIN_GROUP = 'plastron.content_models'
CONTENT_MODEL_CLASSES = entry_points(group=PLUGIN_GROUP)


class ModelClassError(Exception):
    pass


class ModelClassNotFoundError(ModelClassError):
    def __init__(self, model_name: str, *args):
        super().__init__(*args)
        self.model_name = model_name


class ContentModeledResource(RDFResourceBase):
    model_name: str
    is_top_level: bool = False
    HEADER_MAP: Dict = None


def get_model_from_name(model_name: str) -> Type[ContentModeledResource]:
    try:
        return CONTENT_MODEL_CLASSES[model_name].load()
    except KeyError as e:
        raise ModelClassNotFoundError(model_name) from e


def get_model_from_uri(rdf_type: URIRef) -> Type[ContentModeledResource]:
    for plugin in CONTENT_MODEL_CLASSES:
        cls = plugin.load()
        if rdf_type in cls.default_values.get('rdf_type', set()):
            return cls
    raise ModelClassNotFoundError(str(rdf_type))


def guess_model(resource: RDFResource) -> Type[ContentModeledResource]:
    for plugin in CONTENT_MODEL_CLASSES:
        cls = plugin.load()
        if cls.default_values.get('rdf_type', set()) <= set(resource.rdf_type.values):
            return cls
    raise ModelClassError()



"""[ALTO](https://www.loc.gov/standards/alto/) (Analyzed Layout and Text Object) OCR classes"""

from typing import Union, Iterator

# noinspection PyProtectedMember
from lxml.etree import _Element, _ElementTree, QName

from plastron.ocr.core import XYWH, BBox, Scale, RegionBase, OCRError, OCRResource, BlockRegion, LineRegion, WordRegion

XMLNS = {"alto": "http://www.loc.gov/standards/alto/ns-v2#"}


class ALTOResource(OCRResource):
    """ALTO XML OCR resource. The `scale` is determined by using the measurement
    unit given in the ALTO document combined with the image resolution passed to
    the constructor. If no measurement unit can be found, raises an `OCRError`."""

    def __init__(self, doc: _ElementTree, image_resolution: tuple[int, int]):
        self.doc = doc
        try:
            unit = self.doc.xpath('/alto:alto/alto:Description/alto:MeasurementUnit', namespaces=XMLNS)[0].text
        except IndexError:
            raise OCRError('Unable to determine measurement unit from ALTO document')
        self.scale = Scale.from_resolution(image_resolution, unit)

    def get_block_nodes(self):
        """Get the `<TextBlock>` descendant elements of this resource."""
        return self.doc.xpath("//alto:TextBlock", namespaces=XMLNS)

    def get_block_node(self, identifier: str):
        """Get the `<TextBlock>` descendant element whose `@ID` matches
        the given `identifier`."""
        return self.doc.xpath("//alto:TextBlock[@ID=$id]", id=identifier, namespaces=XMLNS)[0]

    def get_block(self, node: _Element) -> 'TextBlock':
        """Get the `TextBlock` object wrapping the given `<TextBlock>` element."""
        return TextBlock(node, self.scale)


class ALTORegion(RegionBase):
    """Region within an `ALTOResource`"""
    def __init__(self, element: _Element, scale: Scale):
        self.element = element
        self.scale = scale
        self.id = self.element.get('ID')
        self._xywh = XYWH(
            x=int(self.element.get('HPOS')),
            y=int(self.element.get('VPOS')),
            w=int(self.element.get('WIDTH')),
            h=int(self.element.get('HEIGHT', 0)),
        ).scale(self.scale)
        self._bbox = BBox.from_xywh(self._xywh)


class TextBlock(ALTORegion, BlockRegion):
    """Wraps an ALTO `<TextBlock>` element."""

    def get_line_nodes(self):
        """Get the `<TextLine>` child elements of this block."""
        return self.element.xpath('alto:TextLine', namespaces=XMLNS)

    def get_line(self, node: _Element):
        """Get a `TextLine` object wrapping the given `<TextLine>` element."""
        return TextLine(node, self.scale)


class TextLine(ALTORegion, LineRegion):
    """Wraps an ALTO `<TextLine>` element."""

    def get_word_nodes(self):
        """Get the `<String>` child elements of this line."""
        return self.element.xpath('alto:String', namespaces=XMLNS)

    def get_word(self, node: _Element) -> 'String':
        """Get a `String` object wrapping the given `<String>` element."""
        return String(node, self.scale)

    @property
    def inlines(self) -> Iterator[Union['String', 'Space', 'Hyphen']]:
        """Iterator over all the inline elements for this line (strings,
        spaces, and hyphens)."""

        for node in self.element.xpath('alto:String|alto:SP|alto:HYP', namespaces=XMLNS):
            tag = QName(node.tag)
            if tag.localname == 'String':
                yield String(node, self.scale)
            elif tag.localname == 'SP':
                yield Space(node, self.scale)
            elif tag.localname == 'HYP':
                yield Hyphen(node, self.scale)


class String(ALTORegion, WordRegion):
    """Wraps an ALTO `<String>` element."""

    def __str__(self):
        return self.content

    def get_content(self) -> str:
        """Get the string content.

        For indexing purposes, if there is a hyphenated word across multiple
        bounding boxes, we create two tokens, each with the full substituted
        content (i.e., the unbroken word) and their respective bounding boxes."""
        if 'SUBS_CONTENT' in self.element.attrib:
            return self.element.get('SUBS_CONTENT')
        else:
            return self.element.get('CONTENT', '')


class Space(ALTORegion):
    def __str__(self):
        return ' '


class Hyphen(ALTORegion):
    def __str__(self):
        return '\N{SOFT HYPHEN}'



""".. include:: ../../../docs/serializers.md"""
import logging

from rdflib import URIRef

from plastron.models.letter import Letter
from plastron.models.newspaper import Issue
from plastron.models.poster import Poster
from plastron.namespaces import bibo, rdf
from plastron.serializers.csv import CSVSerializer
from plastron.serializers.turtle import TurtleSerializer

logger = logging.getLogger(__name__)

SERIALIZER_CLASSES = {
    'text/turtle': TurtleSerializer,
    'turtle': TurtleSerializer,
    'ttl': TurtleSerializer,
    'text/csv': CSVSerializer,
    'csv': CSVSerializer
}

MODEL_MAP = {
    bibo.Image: Poster,
    bibo.Issue: Issue,
    bibo.Letter: Letter
}


def detect_resource_class(graph, subject, fallback=None):
    types = set(graph.objects(URIRef(subject), rdf.type))

    for rdf_type, cls in MODEL_MAP.items():
        if rdf_type in types:
            return cls
    else:
        if fallback is not None:
            return fallback
        else:
            raise RuntimeError(f'Unable to detect resource type for {subject}')



from plastron.namespaces import fedora, xsd
from plastron.rdfmapping.descriptors import ObjectProperty, DataProperty
from plastron.rdfmapping.resources import RDFResource


class FedoraResource(RDFResource):
    created = DataProperty(fedora.created, datatype=xsd.dateTime)
    created_by = DataProperty(fedora.createdBy)
    last_modified = DataProperty(fedora.lastModified, datatype=xsd.dateTime)
    last_modified_by = DataProperty(fedora.lastModifiedBy)
    parent = ObjectProperty(fedora.hasParent)



from PIL import Image

from plastron.namespaces import dcterms, ebucore, fabio, pcdm, pcdmuse, premis
from plastron.rdf import ldp, ore, rdf

# alias the rdflib Namespace
ns = pcdm


@rdf.object_property('members', pcdm.hasMember)
@rdf.object_property('member_of', pcdm.memberOf)
@rdf.object_property('files', pcdm.hasFile)
@rdf.object_property('related', pcdm.hasRelatedObject)
@rdf.object_property('related_of', pcdm.relatedObjectOf)
@rdf.data_property('title', dcterms.title)
@rdf.rdf_class(pcdm.Object)
class Object(ore.Aggregation):
    def add_member(self, obj):
        self.members.append(obj)
        obj.member_of.append(self)

    def add_file(self, obj):
        self.files.append(obj)
        obj.file_of.append(self)

    def add_related(self, obj):
        self.related.append(obj)
        obj.related_of.append(self)

    def gather_files(self, repository):
        for proxy in self.load_proxies(repository):
            page = Object.from_repository(repository, proxy.proxy_for[0])
            for file_uri in page.files:
                file = File.from_repository(repository, file_uri)
                _, graph = repository.get_graph(file_uri)
                file.read(graph)
                yield file

    # recursively create an object and components and that don't yet exist
    def create(self, client, container_path=None, slug=None, headers=None, recursive=True, **kwargs):
        super().create(
            client=client,
            container_path=container_path,
            slug=slug,
            headers=headers,
            recursive=recursive,
            **kwargs
        )
        if recursive:
            client.create_members(self)
            client.create_files(self)
            client.create_related(self)

    def get_new_member(self, rootname, number):
        return Page(title=f'Page {number}', number=number)


@rdf.object_property('file_of', pcdm.fileOf)
@rdf.data_property('mimetype', ebucore.hasMimeType)
@rdf.data_property('filename', ebucore.filename)
@rdf.data_property('size', premis.hasSize)
@rdf.data_property('width', ebucore.width)
@rdf.data_property('height', ebucore.height)
@rdf.object_property('dcmitype', dcterms.type)
@rdf.data_property('title', dcterms.title)
@rdf.rdf_class(pcdm.File)
class File(ldp.NonRdfSource):
    @classmethod
    def from_repository(cls, client, uri, include_server_managed=True):
        obj = super().from_repository(client, uri, include_server_managed)
        obj.source = RepositoryFileSource(uri=uri, client=client)
        return obj

    @classmethod
    def from_source(cls, source=None, **kwargs):
        obj = super().from_source(source=source, **kwargs)
        obj.mimetype = source.mimetype()
        return obj

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        # for image files
        # TODO: move these to a subclass or mix-in?
        self.width = None
        self.height = None

    # upload a binary resource
    def create(self, client, container_path=None, slug=None, headers=None, **kwargs):
        if not client.load_binaries:
            self.logger.info(f'Skipping loading for binary {self.source.filename}')
            return True
        elif self.created:
            return False
        elif self.exists_in_repo(client):
            self.created = True
            return False

        self.logger.info(f'Loading {self.source.filename}')

        if headers is None:
            headers = {}
        headers.update({
            'Content-Type': self.source.mimetype(),
            'Digest': self.source.digest(),
            'Content-Disposition': f'attachment; filename="{self.source.filename}"'
        })

        with self.source as stream:
            super().create(client, container_path=container_path, slug=slug, headers=headers, data=stream, **kwargs)
        self.created = True
        return True

    def update(self, client, recursive=True):
        if not client.load_binaries:
            self.logger.info(f'Skipping update for binary {self.source.filename}')
            return True

        # if this is an image file, see if we can get dimensions
        if self.source.mimetype().startswith('image/'):
            if self.width is None or self.height is None:
                # use PIL
                try:
                    with self.source as stream:
                        with Image.open(stream) as img:
                            self.width = img.width
                            self.height = img.height
                except IOError as e:
                    self.logger.warn(f'Cannot read image file: {e}')

        return super().update(client, recursive=recursive)


@rdf.rdf_class(pcdmuse.PreservationMasterFile)
class PreservationMasterFile(File):
    pass


@rdf.rdf_class(pcdmuse.IntermediateFile)
class IntermediateFile(File):
    pass


@rdf.rdf_class(pcdmuse.ServiceFile)
class ServiceFile(File):
    pass


@rdf.rdf_class(pcdmuse.ExtractedText)
class ExtractedText(File):
    pass


@rdf.rdf_class(pcdm.Collection)
class Collection(Object):
    pass


@rdf.data_property('number', fabio.hasSequenceIdentifier)
@rdf.rdf_class(fabio.Page)
class Page(Object):
    """One page of an item-level resource"""
    pass


FILE_CLASS_FOR = {
    '.tif': PreservationMasterFile,
    '.jpg': IntermediateFile,
    '.txt': ExtractedText,
    '.xml': ExtractedText,
}


def get_file_object(path, source=None):
    extension = path[path.rfind('.'):]
    if extension in FILE_CLASS_FOR:
        cls = FILE_CLASS_FOR[extension]
    else:
        cls = File
    if source is None:
        source = LocalFileSource(path)
    f = cls.from_source(source)
    return f



import os

from rdflib import Graph

from plastron.namespaces import get_manager
from plastron.rdfmapping.resources import RDFResourceBase

nsm = get_manager()


class TurtleSerializer:
    def __init__(self, directory, **_kwargs):
        self.directory_name = directory
        self.content_type = 'text/turtle'
        self.file_extension = '.ttl'
        self.graph = Graph(namespace_manager=nsm)

    def __enter__(self):
        return self

    def write(self, resource: RDFResourceBase, **_kwargs):
        self.graph += resource.graph

    def finish(self):
        with open(os.path.join(self.directory_name, 'metadata.ttl'), mode='wb') as export_file:
            self.graph.serialize(destination=export_file, format='turtle')

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.finish()



from plastron.models.umd import Item
from plastron.models.letter import Letter
from plastron.models.newspaper import Issue
from plastron.models.poster import Poster
from plastron.namespaces import schema
from rdflib import Graph, URIRef, Literal

import pytest

base_uri = 'http://example.com/xyz'


@pytest.mark.parametrize("model_class", [Item, Letter, Poster, Issue])
def test_model_with_copyright_notice(model_class):
    copyright_text = 'Copyright 2024. All rights reserved.'
    model = create_model_with_copyright_notice(model_class, base_uri, copyright_text)
    assert str(model.copyright_notice.value) == copyright_text


@pytest.mark.parametrize("model_class", [Item, Letter, Poster, Issue])
def test_copyright_notice_can_be_set_on_model(model_class):
    copyright_notice = 'Public Domain'

    model = model_class(uri=base_uri)
    model.copyright_notice = Literal(copyright_notice)

    expected = (URIRef(base_uri), URIRef('https://schema.org/copyrightNotice'), Literal(copyright_notice))
    assert expected in model.graph


# Helper Functions

def create_model_with_copyright_notice(model_class, item_uri, copyright_notice_text):
    copyright_notice = f'<> <{schema.copyrightNotice}> \"{copyright_notice_text}\" .'
    model_graph = Graph().parse(data=copyright_notice, format='turtle', publicID=item_uri)
    model = model_class(graph=model_graph, uri=item_uri)

    return model



from plastron.namespaces import ore, dcterms, iana
from plastron.rdfmapping.decorators import rdf_type
from plastron.rdfmapping.descriptors import DataProperty, ObjectProperty
from plastron.rdfmapping.resources import RDFResource, RDFResourceBase


@rdf_type(ore.Proxy)
class Proxy(RDFResource):
    title = DataProperty(dcterms.title)
    prev = ObjectProperty(iana.prev)
    next = ObjectProperty(iana.next, cls='Proxy')
    proxy_for = ObjectProperty(ore.proxyFor)
    proxy_in = ObjectProperty(ore.proxyIn)


class AggregationMixin(RDFResourceBase):
    first = ObjectProperty(iana.first, cls=Proxy)
    last = ObjectProperty(iana.last)



from plastron.models.umd import Item
from plastron.models.letter import Letter
from plastron.models.newspaper import Issue
from plastron.models.poster import Poster
from plastron.namespaces import dcterms
from rdflib import Graph, URIRef

import pytest

base_uri = 'http://example.com/xyz'


@pytest.mark.parametrize("model_class", [Item, Letter, Poster, Issue])
def test_model_with_terms_of_use(model_class):
    model = create_model_with_terms_of_use(
        model_class, base_uri, 'foo'
    )

    assert model.terms_of_use.value == URIRef('http://vocab.lib.umd.edu/termsOfUse#foo')


@pytest.mark.parametrize("model_class", [Item, Letter, Poster, Issue])
def test_model_validation_with_valid_terms_of_use(model_class):
    model = create_model_with_terms_of_use(
        model_class, base_uri, 'test'
    )
    assert model.terms_of_use.is_valid


@pytest.mark.parametrize("model_class", [Item, Letter, Poster, Issue])
def test_model_validation_with_invalid_terms_of_use(model_class):
    model = create_model_with_terms_of_use(
        model_class, base_uri, 'not_valid'
    )
    assert not model.terms_of_use.is_valid


@pytest.mark.parametrize("model_class", [Item, Letter, Poster, Issue])
def test_terms_of_use_can_be_set_on_model(model_class):
    model = model_class(uri=base_uri)
    model.terms_of_use = URIRef('http://vocab.lib.umd.edu/termsOfUse#test')

    expected = (URIRef(base_uri), dcterms.license, URIRef('http://vocab.lib.umd.edu/termsOfUse#test'))
    assert expected in model.graph
    assert model.terms_of_use.is_valid

# Helper Functions


def create_terms_of_use_turtle_format(term_of_use):
    preamble = '@prefix dcterms: <http://purl.org/dc/terms/> .'
    preamble = preamble + '@prefix umdtermsOfUse: <http://vocab.lib.umd.edu/termsOfUse#> .'
    terms_of_use = preamble
    terms_of_use += f'<> dcterms:license umdtermsOfUse:{term_of_use} .'
    return terms_of_use


def create_model_with_terms_of_use(model_class, item_uri, terms_of_use):
    terms_of_use = create_terms_of_use_turtle_format(terms_of_use)

    model_graph = Graph().parse(data=terms_of_use, format='turtle', publicID=item_uri)
    model = model_class(graph=model_graph, uri=item_uri)

    return model



import dataclasses
import logging
from dataclasses import dataclass
from typing import Any

from requests import Session
from requests_jwtauth import HTTPBearerAuth

from plastron.namespaces import dcterms, umdtype
from plastron.rdfmapping.descriptors import DataProperty
from plastron.rdfmapping.resources import RDFResource
from plastron.validation.rules import is_handle

logger = logging.getLogger(__name__)


def parse_handle_string(handle: str) -> list[str]:
    if handle.startswith('hdl:'):
        handle = handle[4:]
    try:
        return handle.split('/', 1)
    except ValueError as e:
        raise HandleError(
            'Handle must be a string in the form "{prefix}/{suffix}" or "hdl:{prefix}/{suffix}'
        ) from e


def parse_result(result: dict[str, Any]) -> dict[str, Any]:
    logger.debug(f'Raw result: {result}')
    if 'request' in result:
        request = result['request']
        del result['request']
        result.update(request)
    return result


@dataclass
class HandleInfo:
    exists: bool
    handle_url: str = None
    prefix: str = None
    suffix: str = None
    url: str = None
    repo: str = None
    repo_id: str = None

    def __str__(self):
        """The handle in `{prefix}/{suffix}` form"""
        return f'{self.prefix}/{self.suffix}'

    @property
    def hdl_uri(self):
        """The handle in `hdl:{prefix}/{suffix}` form"""
        return f'hdl:{self}'


class HandleServiceClient:
    def __init__(self, endpoint_url: str, jwt_token: str, default_prefix: str = None, default_repo: str = None):
        self.endpoint_url = endpoint_url
        self.default_prefix = default_prefix
        self.default_repo = default_repo
        self.session = Session()
        self.session.auth = HTTPBearerAuth(jwt_token)

    def get_info(self, prefix: str, suffix: str):
        url = self.endpoint_url + '/handles/info'
        response = self.session.get(
            url=url,
            params={
                'prefix': prefix,
                'suffix': suffix,
            },
        )
        if not response.ok:
            raise HandleServerError(str(response))

        return HandleInfo(**parse_result(response.json()))

    def find_handle(self, repo_id: str, repo: str = None) -> HandleInfo:
        url = self.endpoint_url + '/handles/exists'
        response = self.session.get(
            url=url,
            params={
                'repo': repo or self.default_repo,
                'repo_id': repo_id,
            },
        )
        if not response.ok:
            raise HandleServerError(str(response))

        return HandleInfo(**parse_result(response.json()))

    def create_handle(self, repo_id: str, url: str, prefix: str = None, repo: str = None) -> HandleInfo:
        request = {
            'prefix': prefix or self.default_prefix,
            'repo': repo or self.default_repo,
            'repo_id': repo_id,
            'url': url,
        }
        response = self.session.post(
            f'{self.endpoint_url}/handles',
            json=request,
        )
        if not response.ok:
            raise HandleServerError(str(response))

        return HandleInfo(exists=True, **parse_result(response.json()))

    def update_handle(self, handle_info: HandleInfo, **fields) -> HandleInfo:
        updated_handle_info = dataclasses.replace(handle_info, **fields)
        response = self.session.patch(
            f'{self.endpoint_url}/handles/{handle_info.prefix}/{handle_info.suffix}',
            json=dataclasses.asdict(updated_handle_info),
        )
        if not response.ok:
            raise HandleServerError(str(response))

        return updated_handle_info


class HandleError(Exception):
    pass


class HandleServerError(HandleError):
    pass


class HandleNotFoundError(HandleServerError):
    def __init__(self, handle, *args):
        super().__init__(*args)
        self.handle = handle


class HandleBearingResource(RDFResource):
    """This class be used by itself for instances where the handle field is the only
    one needed, or it can be used as a mix-in to other full models to give them a handle
    field."""
    handle = DataProperty(dcterms.identifier, datatype=umdtype.handle, validate=is_handle)

    @property
    def has_handle(self) -> bool:
        """Convenience property for whether this object has a valid handle."""
        return bool(len(self.handle) > 0 and self.handle.is_valid)



from plastron.namespaces import ldp
from plastron.rdfmapping.descriptors import ObjectProperty
from plastron.rdfmapping.resources import RDFResource


class LDPContainer(RDFResource):
    contains = ObjectProperty(ldp.contains, repeatable=True)



from plastron.models.ore import AggregationMixin
from plastron.namespaces import pcdm, dcterms, ebucore, premis, xsd
from plastron.rdfmapping.decorators import rdf_type
from plastron.rdfmapping.descriptors import DataProperty, ObjectProperty
from plastron.rdfmapping.resources import RDFResource


@rdf_type(pcdm.Object)
class PCDMObject(RDFResource, AggregationMixin):
    title = DataProperty(dcterms.title)
    has_member = ObjectProperty(pcdm.hasMember, repeatable=True, cls='PCDMObject')
    member_of = ObjectProperty(pcdm.memberOf, repeatable=True)
    has_file = ObjectProperty(pcdm.hasFile, repeatable=True, cls='PCDMFile')


@rdf_type(pcdm.File)
class PCDMFile(RDFResource):
    title = DataProperty(dcterms.title)
    file_of = ObjectProperty(pcdm.fileOf, repeatable=True)
    mime_type = DataProperty(ebucore.hasMimeType)
    filename = DataProperty(ebucore.filename)
    size = DataProperty(premis.hasSize, datatype=xsd.long)
    checksum = ObjectProperty(premis.hasMessageDigest)

    def __str__(self):
        return str(self.title or self.uri)


class PCDMImageFile(PCDMFile):
    height = DataProperty(ebucore.height)
    width = DataProperty(ebucore.width)



from plastron.models.authorities import UMD_RIGHTS_STATEMENTS
from plastron.models.newspaper import Issue

base_uri = 'http://example.com/xyz'


def test_issue_invalid_with_no_fields():
    issue = Issue()
    assert not issue.is_valid


def test_issue_valid_with_only_required_fields():
    issue = Issue()

    # Only provide required fields
    issue.identifier = 'test_issue'
    issue.title = 'Test Issue'
    issue.date = '1970-01-01'
    issue.volume = '1'
    issue.issue = '1'
    issue.edition = '1'
    issue.rights = list(UMD_RIGHTS_STATEMENTS.keys())[0]

    assert issue.is_valid


def test_issue_invalid_with_bad_rights_statement():
    issue = Issue()

    # Only provide required fields
    issue.identifier = 'test_issue'
    issue.title = 'Test Issue'
    issue.date = '1970-01-01'
    issue.volume = '1'
    issue.issue = '1'
    issue.edition = '1'

    issue.rights = 'not a rights statement vocab URL'

    assert not issue.is_valid



from rdflib import URIRef

from plastron.handles import HandleBearingResource
from plastron.models import ContentModeledResource
from plastron.models.authorities import UMD_TERMS_OF_USE_STATEMENTS, UMD_PRESENTATION_SETS
from plastron.models.fedora import FedoraResource
from plastron.models.pcdm import PCDMObject
from plastron.models.page import Page
from plastron.namespaces import bibo, dc, dcmitype, dcterms, edm, geo, rel, skos, ore, owl, umd, schema, pcdm
from plastron.rdfmapping.decorators import rdf_type
from plastron.rdfmapping.descriptors import ObjectProperty, DataProperty
from plastron.rdfmapping.resources import RDFResource
from plastron.validation.rules import is_edtf_formatted
from plastron.validation.vocabularies import ControlledVocabularyProperty


class AuthorityRecord(RDFResource):
    same_as = ObjectProperty(owl.sameAs, repeatable=True)


@rdf_type(edm.Agent)
class Agent(AuthorityRecord):
    pass


@rdf_type(skos.Concept)
class Concept(AuthorityRecord):
    pass


@rdf_type(edm.Place)
class Place(AuthorityRecord):
    lat = DataProperty(geo.lat)
    long = DataProperty(geo.long)


@rdf_type(dcmitype.Collection)
class Collection(AuthorityRecord):
    pass


@rdf_type(bibo.Letter, umd.Letter)
class Letter(ContentModeledResource, PCDMObject, HandleBearingResource, FedoraResource):
    model_name = 'Letter'
    is_top_level = True

    member_of = ObjectProperty(pcdm.memberOf)
    has_member = ObjectProperty(pcdm.hasMember, repeatable=True, cls=Page)
    title = DataProperty(dcterms.title, required=True)
    author = ObjectProperty(rel.aut, repeatable=True, embed=True, cls=Agent)
    recipient = ObjectProperty(bibo.recipient, repeatable=True, embed=True, cls=Agent)
    part_of = ObjectProperty(dcterms.isPartOf, required=True, embed=True, cls=Collection)
    place = ObjectProperty(dcterms.spatial, repeatable=True, embed=True, cls=Place)
    subject = ObjectProperty(dcterms.subject, repeatable=True, embed=True, cls=Concept)
    rights = ObjectProperty(dcterms.rights, required=True)
    copyright_notice = DataProperty(schema.copyrightNotice)
    identifier = DataProperty(dcterms.identifier, required=True)
    type = DataProperty(edm.hasType, required=True)
    date = DataProperty(dc.date, datatype=URIRef('http://id.loc.gov/datatypes/edtf'), validate=is_edtf_formatted)
    language = DataProperty(dc.language, required=True)
    description = DataProperty(dcterms.description, required=True)
    bibliographic_citation = DataProperty(dcterms.bibliographicCitation, required=True)
    extent = DataProperty(dcterms.extent, required=True)
    rights_holder = DataProperty(dcterms.rightsHolder, required=True)
    terms_of_use = ControlledVocabularyProperty(dcterms.license, vocab=UMD_TERMS_OF_USE_STATEMENTS)
    presentation_set = ControlledVocabularyProperty(ore.isAggregatedBy, repeatable=True, vocab=UMD_PRESENTATION_SETS)

    HEADER_MAP = {
        'title': 'Title',
        'rights_holder': 'Rights Holder',
        'extent': 'Extent',
        'bibliographic_citation': 'Bibliographic Citation',
        'description': 'Description',
        'language': 'Language',
        'date': 'Date',
        'type': 'Resource Type',
        'rights': 'Rights',
        'terms_of_use': 'Terms of Use',
        'copyright_notice': 'Copyright Notice',
        'subject': {
            'label': 'Subject',
        },
        'place': {
            'label': 'Location',
            'long': 'Longitude',
            'lat': 'Latitude',
        },
        'part_of': {
            'label': 'Archival Collection',
            'same_as': 'Handle/Link',
        },
        'identifier': 'Identifier',
        'recipient': {
            'label': 'Recipient',
        },
        'author': {
            'label': 'Author',
        },
        'handle': 'Handle',
        'presentation_set': 'Presentation Set',
    }



from rdflib import URIRef

from plastron.handles import HandleBearingResource
from plastron.models import ContentModeledResource
from plastron.models.authorities import UMD_PRESENTATION_SETS, UMD_TERMS_OF_USE_STATEMENTS
from plastron.models.fedora import FedoraResource
from plastron.models.pcdm import PCDMObject
from plastron.models.page import Page
from plastron.namespaces import bibo, dcterms, dc, edm, geo, ore, schema, umd, xsd, pcdm
from plastron.rdfmapping.decorators import rdf_type
from plastron.rdfmapping.descriptors import ObjectProperty, DataProperty, Property
from plastron.validation.rules import is_edtf_formatted
from plastron.validation.vocabularies import ControlledVocabularyProperty


@rdf_type(bibo.Image, umd.Poster)
class Poster(ContentModeledResource, PCDMObject, HandleBearingResource, FedoraResource):
    model_name = 'Poster'
    is_top_level = True

    member_of = ObjectProperty(pcdm.memberOf)
    has_member = ObjectProperty(pcdm.hasMember, repeatable=True, cls=Page)
    title = DataProperty(dcterms.title, required=True)
    alternative = DataProperty(dcterms.alternative)
    publisher = DataProperty(dc.publisher)
    part_of = DataProperty(dcterms.isPartOf, required=True)
    # use the full URI since "dc.format" returns the format method of Namespace
    format = DataProperty(URIRef('http://purl.org/dc/elements/1.1/format'), required=True)
    # this is a Property since the extant data in fcrepo contains both URIs and literals,
    # so neither ObjectProperty nor DataProperty would map all values correctly
    type = Property(edm.hasType, required=True)
    date = DataProperty(dc.date, datatype=xsd.date, validate=is_edtf_formatted)
    language = DataProperty(dc.language, required=True)
    description = DataProperty(dcterms.description)
    extent = DataProperty(dcterms.extent)
    issue = DataProperty(bibo.issue)
    locator = DataProperty(bibo.locator, required=True)
    location = DataProperty(dc.coverage)
    longitude = DataProperty(geo.long)
    latitude = DataProperty(geo.lat)
    subject = DataProperty(dc.subject, repeatable=True)
    rights = ObjectProperty(dcterms.rights, required=True)
    terms_of_use = ControlledVocabularyProperty(dcterms.license, vocab=UMD_TERMS_OF_USE_STATEMENTS)
    copyright_notice = DataProperty(schema.copyrightNotice)
    identifier = DataProperty(dcterms.identifier, required=True)
    place = ObjectProperty(dcterms.spatial)
    presentation_set = ControlledVocabularyProperty(ore.isAggregatedBy, repeatable=True, vocab=UMD_PRESENTATION_SETS)

    HEADER_MAP = {
        'title': 'Title',
        'alternative': 'Alternate Title',
        'publisher': 'Publisher',
        'part_of': 'Collection',
        'format': 'Format',
        'type': 'Resource Type',
        'date': 'Date',
        'language': 'Language',
        'description': 'Description',
        'extent': 'Extent',
        'issue': 'Issue',
        'locator': 'Identifier/Call Number',
        'location': 'Location',
        'longitude': 'Longitude',
        'latitude': 'Latitude',
        'subject': 'Subject',
        'rights': 'Rights',
        'terms_of_use': 'Terms of Use',
        'copyright_notice': 'Copyright Notice',
        'identifier': 'Identifier',
        'handle': 'Handle',
        'presentation_set': 'Presentation Set',
    }



from rdflib import RDF

from plastron.namespaces import dcterms, oa, prov
from plastron.rdf import ldp, rdf

# alias the rdflib Namespace
ns = oa


# Annotation resources
@rdf.object_property('body', oa.hasBody, embed=True)
@rdf.object_property('target', oa.hasTarget, embed=True)
@rdf.object_property('motivation', oa.motivatedBy)
@rdf.rdf_class(oa.Annotation)
class Annotation(ldp.Resource):
    def __str__(self):
        return ' '.join([str(body) for body in self.body])

    def add_body(self, body):
        self.body.append(body)
        body.annotation = self

    def add_target(self, target):
        self.target.append(target)
        target.annotation = self


@rdf.data_property('value', RDF.value)
@rdf.data_property('content_type', dcterms['format'])
@rdf.rdf_class(oa.TextualBody)
class TextualBody(ldp.Resource):
    def __str__(self):
        value = str(self.value)
        return value if len(value) <= 25 else value[:24] + '…'


@rdf.object_property('selector', oa.hasSelector, embed=True)
@rdf.object_property('source', oa.hasSource)
@rdf.rdf_class(oa.SpecificResource)
class SpecificResource(ldp.Resource):
    def __str__(self):
        return ' '.join([str(selector) for selector in self.selector])

    def add_selector(self, selector):
        self.selector.append(selector)
        selector.annotation = self


@rdf.data_property('value', RDF.value)
@rdf.object_property('conforms_to', dcterms.conformsTo)
@rdf.rdf_class(oa.FragmentSelector)
class FragmentSelector(ldp.Resource):
    def __str__(self):
        return str(self.value)


@rdf.data_property('value', RDF.value)
@rdf.rdf_class(oa.XPathSelector)
class XPathSelector(ldp.Resource):
    def __str__(self):
        return str(self.value)


@rdf.object_property('derived_from', prov.wasDerivedFrom)
class FullTextAnnotation(Annotation):
    pass



import plastron.validation.rules
from plastron.namespaces import edm, geo, rdfs, owl
from plastron.rdf import ldp, rdf


def create_authority(graph, subject):
    if (subject, rdf.ns.type, edm.Place) in graph:
        return Place.from_graph(graph, subject)
    else:
        return LabeledThing.from_graph(graph, subject)


@rdf.data_property('label', rdfs.label)
@rdf.object_property('same_as', owl.sameAs)
class LabeledThing(ldp.Resource):
    VALIDATION_RULESET = {
        'label': {
            'required': True
        },
        'same_as': {}
    }

    def validate(self, parent_prop, result):
        '''
        Validates this object based on its internal ruleset, populating
        the provided result object with the result of the validation.
        '''
        ruleset = self.VALIDATION_RULESET
        for field, rules in ruleset.items():
            for rule_name, arg in rules.items():
                rule = getattr(plastron.validation.rules, rule_name)
                prop = getattr(self, field)
                if rule(prop, arg):
                    result.passes(parent_prop, rule, arg)
                else:
                    result.fails(parent_prop, rule, arg)


@rdf.data_property('lat', geo.lat)
@rdf.data_property('lon', geo.long)
class Place(LabeledThing):
    pass



from rdflib import RDF, URIRef, Literal

from plastron.namespaces import dcterms, oa, prov, sc
from plastron.rdfmapping.decorators import rdf_type
from plastron.rdfmapping.descriptors import ObjectProperty, DataProperty
from plastron.rdfmapping.embed import embedded
from plastron.rdfmapping.resources import RDFResource

# alias the rdflib Namespace
ns = oa


# Annotation resources
@rdf_type(oa.Annotation)
class Annotation(RDFResource):
    body = ObjectProperty(oa.hasBody, repeatable=True, cls=RDFResource)
    target = ObjectProperty(oa.hasTarget, cls=RDFResource)
    motivation = ObjectProperty(oa.motivatedBy)

    def __str__(self):
        return ' '.join([str(body) for body in self.body])

    def add_body(self, body):
        self.body.append(body)
        body.annotation = self

    def add_target(self, target):
        self.target.append(target)
        target.annotation = self


@rdf_type(oa.TextualBody)
class TextualBody(RDFResource):
    value = DataProperty(RDF.value)
    content_type = DataProperty(dcterms['format'])

    def __str__(self):
        value = str(self.value)
        return value if len(value) <= 25 else value[:24] + '…'


@rdf_type(oa.SpecificResource)
class SpecificResource(RDFResource):
    selector = ObjectProperty(oa.hasSelector, cls=RDFResource)
    source = ObjectProperty(oa.hasSource)

    def __str__(self):
        return ' '.join([str(selector) for selector in self.selector])

    def add_selector(self, selector):
        self.selector.append(selector)
        selector.annotation = self


@rdf_type(oa.FragmentSelector)
class FragmentSelector(RDFResource):
    value = DataProperty(RDF.value)
    conforms_to = ObjectProperty(dcterms.conformsTo)

    def __str__(self):
        return str(self.value)


@rdf_type(oa.XPathSelector)
class XPathSelector(RDFResource):
    value = DataProperty(RDF.value)

    def __str__(self):
        return str(self.value)


class FullTextAnnotation(Annotation):
    derived_from = ObjectProperty(prov.wasDerivedFrom)


class TextblockOnPage(Annotation):
    derived_from = ObjectProperty(prov.wasDerivedFrom, cls=RDFResource)

    @classmethod
    def from_textblock(cls, textblock, page, scale, ocr_file):
        xywh = ','.join([str(i) for i in textblock.xywh(scale)])
        return cls(
            body=embedded(TextualBody)(
                value=textblock.text(scale=scale),
                content_type='text/plain'
            ),
            target=embedded(SpecificResource)(
                source=URIRef(page.url),
                selector=embedded(FragmentSelector)(
                    value=Literal(f'xywh={xywh}'),
                    conforms_to=URIRef('http://www.w3.org/TR/media-frags/'),
                ),
            ),
            derived_from=embedded(SpecificResource)(
                source=URIRef(ocr_file.url),
                selector=embedded(XPathSelector)(value=f'//*[@ID="{textblock.id}"]'),
            ),
            motivation=sc.painting
        )



from lxml.etree import parse, XMLSyntaxError

from plastron.handles import HandleBearingResource
from plastron.models import ContentModeledResource
from plastron.models.annotations import TextblockOnPage
from plastron.models.authorities import UMD_TERMS_OF_USE_STATEMENTS, UMD_PRESENTATION_SETS, UMD_RIGHTS_STATEMENTS, Agent
from plastron.models.fedora import FedoraResource
from plastron.models.page import Page, File
from plastron.models.pcdm import PCDMObject, PCDMFile
from plastron.namespaces import bibo, carriers, dc, dcterms, fabio, ndnp, ore, pcdm, pcdmuse, schema, umd
from plastron.ocr.alto import ALTOResource
from plastron.rdfmapping.decorators import rdf_type
from plastron.rdfmapping.descriptors import DataProperty, ObjectProperty
from plastron.validation.rules import is_iso_8601_date
from plastron.validation.vocabularies import ControlledVocabularyProperty


@rdf_type(bibo.Issue, umd.Newspaper)
class Issue(ContentModeledResource, PCDMObject, HandleBearingResource, FedoraResource):
    """Newspaper issue"""
    model_name = 'Issue'
    is_top_level = True

    member_of = ObjectProperty(pcdm.memberOf)
    has_member = ObjectProperty(pcdm.hasMember, repeatable=True, cls=Page)
    has_file = ObjectProperty(pcdm.hasFile, repeatable=True, cls=File)
    title = DataProperty(dcterms.title, required=True)
    date = DataProperty(dc.date, required=True, validate=is_iso_8601_date)
    volume = DataProperty(bibo.volume, required=True)
    issue = DataProperty(bibo.issue, required=True)
    edition = DataProperty(bibo.edition, required=True)
    presentation_set = ControlledVocabularyProperty(ore.isAggregatedBy, repeatable=True, vocab=UMD_PRESENTATION_SETS)
    copyright_notice = DataProperty(schema.copyrightNotice)
    terms_of_use = ControlledVocabularyProperty(dcterms.license, vocab=UMD_TERMS_OF_USE_STATEMENTS)
    rights = ControlledVocabularyProperty(dcterms.rights, required=True, vocab=UMD_RIGHTS_STATEMENTS)
    rights_holder = ObjectProperty(dcterms.rightsHolder, repeatable=True, embed=True, cls=Agent)

    HEADER_MAP = {
        'title': 'Title',
        'date': 'Date',
        'volume': 'Volume',
        'issue': 'Issue',
        'edition': 'Edition',
        'handle': 'Handle',
        'presentation_set': 'Presentation Set',
        'copyright_notice': 'Copyright Notice',
        'terms_of_use': 'Terms of Use',
        'rights': 'Rights Statement',
        'rights_holder': {
            'label': 'Rights Holder',
        },
    }


@rdf_type(fabio.Metadata)
class IssueMetadata(PCDMObject):
    """Additional metadata about an issue"""
    pass


@rdf_type(fabio.MetadataDocument)
class MetadataFile(PCDMFile):
    """A binary file containing metadata in non-RDF formats (METS, MODS, etc.)"""
    pass


@rdf_type(ndnp.Page)
class Page(PCDMObject):
    """Newspaper page"""
    issue = ObjectProperty(pcdm.memberOf, cls=Issue)
    number = DataProperty(ndnp.number)
    frame = DataProperty(ndnp.sequence)

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.ordered = True
        self.ocr = None
        self.ocr_file = None

    def parse_ocr(self):
        # try to get an OCR file
        # if there isn't one, just skip it
        try:
            ocr_file = next(self.files_for('ocr'))
        except StopIteration:
            return

        # load ALTO XML into page object, for text extraction
        try:
            with ocr_file.source as stream:
                tree = parse(stream)
        except OSError:
            raise RuntimeError("Unable to read {0}".format(ocr_file.filename))
        except XMLSyntaxError:
            raise RuntimeError("Unable to parse {0} as XML".format(ocr_file.filename))

        # read in resolution from issue METS data
        master = next(self.files_for('master'))
        self.ocr_file = ocr_file
        self.ocr = ALTOResource(tree, master.resolution)

    def textblocks(self):
        if self.ocr is None:
            raise StopIteration()
        # extract text blocks from ALTO XML for this page
        for textblock in self.ocr.textblocks():
            yield TextblockOnPage.from_textblock(textblock, page=self, scale=self.ocr.scale, ocr_file=self.ocr_file)

    def files_for(self, use):
        for f in self.files:
            if f.use == use:
                yield f


class File(PCDMFile):
    """Newspaper file"""

    @classmethod
    def from_repository(cls, client, uri, include_server_managed=True):
        obj = super().from_repository(client, uri, include_server_managed)

        types = obj.rdf_type.values
        if pcdmuse.PreservationMasterFile in types:
            obj.use = 'master'
            # TODO: how to not hardcode this?
            obj.resolution = (400, 400)
        elif pcdmuse.IntermediateFile in types:
            obj.use = 'service'
        elif pcdmuse.ServiceFile in types:
            obj.use = 'derivative'
        elif pcdmuse.ExtractedText in types:
            obj.use = 'ocr'

        return obj


@rdf_type(bibo.Article)
class Article(PCDMObject):
    """Newspaper article"""
    issue = ObjectProperty(pcdm.memberOf, cls=Issue)
    start_page = DataProperty(bibo.pageStart)
    end_page = DataProperty(bibo.pageEnd)


@rdf_type(carriers.hd)
class Reel(PCDMObject):
    """NDNP reel is an ordered sequence of frames"""
    id = DataProperty(dcterms.identifier)

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.sequence_attr = ('Frame', 'frame')



from plastron.models.umd import Item
from plastron.models.letter import Letter
from plastron.models.newspaper import Issue
from plastron.models.poster import Poster
from plastron.namespaces import ore
from rdflib import Graph, URIRef

import pytest

base_uri = 'http://example.com/xyz'


@pytest.mark.parametrize("model_class", [Item, Letter, Poster, Issue])
def test_model_with_single_presentation_set(model_class):
    model = create_model_with_presentation_set(
        model_class, base_uri, ['foo']
    )

    assert len(model.presentation_set) == 1
    assert model.presentation_set.value == URIRef('http://vocab.lib.umd.edu/set#foo')


@pytest.mark.parametrize("model_class", [Item, Letter, Poster, Issue])
def test_model_validation_with_valid_presentation_set(model_class):
    model = create_model_with_presentation_set(
        model_class, base_uri, ['test']
    )
    assert model.presentation_set.is_valid


@pytest.mark.parametrize("model_class", [Item, Letter, Poster, Issue])
def test_model_validation_with_invalid_presentation_set(model_class):
    model = create_model_with_presentation_set(
        model_class, base_uri, ['not_valid']
    )
    assert not model.presentation_set.is_valid


@pytest.mark.parametrize("model_class", [Item, Letter, Poster, Issue])
def test_model_with_multiple_presentation_sets(model_class):
    model = create_model_with_presentation_set(
        model_class, base_uri, ['foo', 'bar']
    )

    assert len(model.presentation_set) == 2

    expected = sorted([URIRef('http://vocab.lib.umd.edu/set#bar'), URIRef('http://vocab.lib.umd.edu/set#foo')])
    assert sorted(list(model.presentation_set.values)) == expected


@pytest.mark.parametrize("model_class", [Item, Letter, Poster, Issue])
def test_single_presentation_set_can_be_set_on_model(model_class):
    model = model_class(uri=base_uri)
    model.presentation_set = URIRef('http://vocab.lib.umd.edu/set#foobar')

    expected = (URIRef(base_uri), ore.isAggregatedBy, URIRef('http://vocab.lib.umd.edu/set#foobar'))
    assert expected in model.graph


@pytest.mark.parametrize("model_class", [Item, Letter, Poster, Issue])
def test_multiple_presentation_sets_can_be_set_on_model(model_class):
    base_uri_ref = URIRef(base_uri)

    model = model_class(uri=base_uri)
    model.presentation_set.add(URIRef('http://vocab.lib.umd.edu/set#foobar'))
    model.presentation_set.add(URIRef('http://vocab.lib.umd.edu/set#barbaz'))

    expected = [
        (base_uri_ref, ore.isAggregatedBy, URIRef('http://vocab.lib.umd.edu/set#foobar')),
        (base_uri_ref, ore.isAggregatedBy, URIRef('http://vocab.lib.umd.edu/set#barbaz'))
    ]

    for e in expected:
        assert e in model.graph


# Helper Functions

def create_presentation_set_turtle_format(set_names):
    preamble = '@prefix ore: <http://www.openarchives.org/ore/terms/> .'
    preamble = preamble + '@prefix umdset: <http://vocab.lib.umd.edu/set#> .'
    presentation_set = preamble
    for set_name in set_names:
        presentation_set += f'<> ore:isAggregatedBy umdset:{set_name} .'
    return presentation_set


def create_model_with_presentation_set(model_class, item_uri, set_names):
    presentation_set = create_presentation_set_turtle_format(set_names)
    model_graph = Graph().parse(data=presentation_set, format='turtle', publicID=item_uri)
    model = model_class(graph=model_graph, uri=item_uri)

    return model



from plastron.models import ContentModeledResource
from plastron.models.pcdm import PCDMObject, PCDMFile
from plastron.namespaces import fabio, pcdm, dcterms
from plastron.rdfmapping.decorators import rdf_type
from plastron.rdfmapping.descriptors import ObjectProperty, DataProperty


class File(ContentModeledResource, PCDMFile):
    model_name = 'File'
    is_top_level = False

    file_of = ObjectProperty(pcdm.fileOf)


@rdf_type(fabio.Page)
class Page(ContentModeledResource, PCDMObject):
    model_name = 'Page'
    is_top_level = False

    member_of = ObjectProperty(pcdm.memberOf)
    has_file = ObjectProperty(pcdm.hasFile, repeatable=True, cls=File)
    title = DataProperty(dcterms.title)
    number = DataProperty(fabio.hasSequenceIdentifier)

    def __str__(self):
        return str(self.title or self.uri)



from typing import NamedTuple, Iterator, TypeVar, Optional, Iterable

# noinspection PyProtectedMember
from lxml.etree import _Element


class XYWH(NamedTuple):
    """Rectangular region defined by its top-left corner (x, y) coordinates
    plus its width and height (w, h).

    ```pycon
    >>> from plastron.ocr.core import XYWH

    >>> region = XYWH(100, 120, 50, 60)

    >>> str(region)
    '100,120,50,60'

    ```
    """
    x: int
    """X-axis coordinate of the top-left corner of the region"""
    y: int
    """Y-axis coordinate of the top-left corner of the region"""
    w: int
    """Width of the region (i.e., size along the X-axis)"""
    h: int
    """Height of the region (i.e., size along the Y-axis)"""

    @classmethod
    def from_bbox(cls, bbox: 'BBox'):
        """Creates an `XYWH` object representing the same region as the given
        `BBox` object."""
        return cls(x=bbox.x1, y=bbox.y1, w=bbox.x2 - bbox.x1, h=bbox.y2 - bbox.y1)

    def __str__(self):
        return ','.join(map(str, (self.x, self.y, self.w, self.h)))

    def scale(self, scale: 'Scale') -> 'XYWH':
        """Return a new `XYWH` with coordinates adjusted by the given `Scale`."""
        return XYWH(
            x=round(self.x * scale.x),
            y=round(self.y * scale.y),
            w=round(self.w * scale.x),
            h=round(self.h * scale.y),
        )


class BBox(NamedTuple):
    """Rectangular region defined by its top-left (x1, y1) and bottom-right
    (x2, y2) coordinates.

    ```pycon
    >>> from plastron.ocr.core import BBox

    >>> region = BBox(100, 120, 150, 180)

    >>> str(region)
    '100,120,150,180'

    ```
    """
    x1: int
    """X-axis coordinate of the top-left corner of the region"""
    y1: int
    """Y-axis coordinate of the top-left corner of the region"""
    x2: int
    """X-axis coordinate of the bottom-right corner of the region"""
    y2: int
    """Y-axis coordinate of the bottom-right corner of the region"""

    @classmethod
    def from_xywh(cls, xywh: XYWH):
        """Creates a `BBox` object representing the same region as the given
        `XYWH` object."""
        return cls(x1=xywh.x, y1=xywh.y, x2=xywh.x + xywh.w, y2=xywh.y + xywh.h)

    def __str__(self):
        return ','.join(map(str, (self.x1, self.y1, self.x2, self.y2)))

    def scale(self, scale: 'Scale') -> 'BBox':
        """Return a new `BBox` with coordinates adjusted by the given `Scale`."""
        return BBox(
            x1=round(self.x1 * scale.x),
            y1=round(self.y1 * scale.y),
            x2=round(self.x2 * scale.x),
            y2=round(self.y2 * scale.y),
        )


class Scale(NamedTuple):
    """Conversion factor between a given measurement unit and pixels, given the
    image resolution in pixels per inch (a.k.a. DPI). Supported measurement units
    are:

    * `inch1200` (1/1200 of an inch)
    * `mm10` (1/10 of a millimeter)
    * `pixel`
    """
    x: float
    """Horizontal scaling factor"""
    y: float
    """Vertical scaling factor"""

    @classmethod
    def from_resolution(cls, image_resolution: tuple[int, int], unit: str = 'pixel'):
        xres = image_resolution[0]
        yres = image_resolution[1]
        if unit == 'inch1200':
            return cls(xres / 1200.0, yres / 1200.0)
        elif unit == 'mm10':
            return cls(xres / 254.0, yres / 254.0)
        elif unit == 'pixel':
            return cls(1, 1)
        else:
            raise ValueError(f"Unknown MeasurementUnit '{unit}'")


B = TypeVar('B', bound='BlockRegion')
"""`BlockRegion` type"""

L = TypeVar('L', bound='LineRegion')
"""`LineRegion` type"""

W = TypeVar('W', bound='WordRegion')
"""`WordRegion` type"""


class OCRResource:
    """Base class for classes representing complete OCR XML files.

    Conceptually, an OCRResource has the following abstract structure:

    * Resource has 0-or-more Blocks
    * Block has 0-or-more Lines
    * Line has 0-or-more Words

    Specific concrete classes may have other data structures not covered
    here.

    Subclasses must implement three methods:

    * `get_block_nodes()`
    * `get_block_node()`
    * `get_block()`
    """
    def __iter__(self) -> Iterator[B]:
        return self.blocks

    @property
    def blocks(self) -> Iterator[B]:
        """Iterator over the blocks for this resource."""
        for node in self.get_block_nodes():
            yield self.get_block(node)

    def block(self, identifier: str) -> Optional[B]:
        """Retrieve an individual block by identifier."""
        try:
            return self.get_block(self.get_block_node(identifier))
        except IndexError:
            return None

    def get_block_nodes(self) -> Iterable[_Element]:
        """Return an iterable of XML elements representing block-level
        subdivisions of the OCR resource.

        *Must be implemented by subclasses.*"""
        raise NotImplementedError

    def get_block_node(self, identifier: str) -> _Element:
        """Return a single XML element, selected by the `identifier`.

        *Must be implemented by subclasses.*"""
        raise NotImplementedError

    def get_block(self, node: _Element) -> B:
        """Return a `BlockRegion` subclass that encapsulates the
        given XML element `node`.

        *Must be implemented by subclasses.*"""
        raise NotImplementedError

    def words(self) -> Iterator[W]:
        """Iterator over all `WordRegion` elements contained in
        this OCR resource."""
        for block in self.blocks:
            for line in block.lines():
                for word in line.words():
                    yield word


class RegionBase:
    """Base class for classes representing regions of a page with
    a specified bounding box. Subclasses should populate the `_xywh`
    and `_bbox` attributes used to back the `xywh` and `bbox`
    properties."""

    _xywh: XYWH
    _bbox: BBox

    @property
    def xywh(self):
        """Bounding box as `XYWH` (X, Y, Width, Height) coordinates."""
        return self._xywh

    @property
    def bbox(self):
        """Bounding box as `BBox` (Top-left and bottom-right X, Y) coordinates."""
        return self._bbox


class BlockRegion(RegionBase):
    """Base class for classes representing a block-like region of
    a page (e.g., ALTO `TextBlock` or hOCR `ocr_carea`)."""

    def __iter__(self):
        return self.lines

    def __str__(self):
        return ''.join(str(s) + '\n' for s in self.lines())

    def lines(self) -> Iterator[L]:
        """Iterator returning the content of this block region
        as individual lines."""
        for node in self.get_line_nodes():
            yield self.get_line(node)

    def get_line_nodes(self) -> Iterable[_Element]:
        """Return an iterable of XML elements representing the
        lines within this block.

        *Must be implemented by subclasses.*"""
        raise NotImplementedError

    def get_line(self, node: _Element) -> L:
        """Return a `LineRegion` subclass that encapsulates the
        given XML element `node`.

        *Must be implemented by subclasses.*"""
        raise NotImplementedError


class LineRegion(RegionBase):
    """Base class for classes representing a single line in an OCR
    resource (e.g., ALTO `TextLine` or hOCR `ocr_line`)."""
    def __str__(self):
        return ' '.join(str(w) for w in self.words())

    def __iter__(self):
        return self.words

    def words(self) -> Iterator[W]:
        """Iterator returning the content of this line as individual
        words."""
        for node in self.get_word_nodes():
            yield self.get_word(node)

    def get_word_nodes(self) -> Iterable[_Element]:
        """Return an iterable of XML elements representing the
        words within this line.

        *Must be implemented by subclasses.*"""
        raise NotImplementedError

    def get_word(self, node: _Element) -> W:
        """Return a `WordRegion` subclass that encapsulates the
        given XML element `node`.

        *Must be implemented by subclasses.*"""
        raise NotImplementedError


class WordRegion(RegionBase):
    """Base class for classes representing a single word within the
    OCR resource."""
    def __str__(self):
        return self.content

    @property
    def content(self) -> str:
        """The string content of this word."""
        return self.get_content()

    def get_content(self) -> str:
        """Return the string content of this word.

        *Must be implemented by subclasses.*"""
        raise NotImplementedError


class OCRError(Exception):
    """OCR-related error"""
    pass


class OCRFileError(OCRError):
    pass


class ImageFileError(OCRError):
    pass


class UnrecognizedOCRFormatError(OCRError):
    pass



from plastron.client import Client
from plastron.namespaces import dcterms, iana, ore
from plastron.rdf import ldp, rdf

# alias the rdflib Namespace
ns = ore


@rdf.object_property('proxy_for', ore.proxyFor)
@rdf.object_property('proxy_in', ore.proxyIn)
@rdf.data_property('title', dcterms.title)
@rdf.rdf_class(ore.Proxy)
class Proxy(ldp.Resource):
    pass


# add these after the class definition, so Proxy can refer to itself
Proxy.add_object_property('prev', iana.prev, obj_class=Proxy)
Proxy.add_object_property('next', iana.next, obj_class=Proxy)


@rdf.object_property('first', iana.first, obj_class=Proxy)
@rdf.object_property('last', iana.last, obj_class=Proxy)
class Aggregation(ldp.Resource):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.first = []
        self.last = []

    def append(self, proxy):
        if len(self.first) == 0:
            self.first = proxy
        else:
            last = self.last.values[0]
            # update the current last item to point to the new item
            last.next = proxy
            proxy.prev = last
        # new item is always the last in the list
        self.last = proxy
        return proxy

    def append_proxy(self, proxy_for, title=None):
        if title is None:
            title = f'Proxy for {proxy_for} in {self}'
        return self.append(Proxy(
            title=title,
            proxy_for=proxy_for,
            proxy_in=self
        ))

    def load_proxies(self, client: Client):
        """
        Returns an AggregationIterator over the resources in this aggregation,
        fully filled in with data from the given repository.
        """
        return AggregationIterator(self, client)

    def __iter__(self):
        """
        Returns an AggregationIterator over the resources in this aggregation,
        without any further calls to a repository.
        """
        return AggregationIterator(self)

    def proxies(self):
        return [proxy for proxy in self]

    def create(self, client: Client, container_path=None, slug=None, headers=None, recursive=True, **kwargs):
        super().create(
            client=client,
            container_path=container_path,
            slug=slug,
            headers=headers,
            recursive=recursive,
            **kwargs
        )
        if recursive:
            client.create_proxies(self)


class AggregationIterator:
    """
    Iterator over the resources in an aggregation. If a repository is
    specified, then before each resource is returned, its metadata is
    retrieved from that repository.
    """
    def __init__(self, aggregation: Aggregation, client: Client = None):
        self.aggregation = aggregation
        self.repository = client
        if len(self.aggregation.first) == 0:
            # no members of the aggregation
            self.next_proxy = None
        else:
            self.next_proxy = self.aggregation.first.values[0]

    def __iter__(self):
        return self

    def __next__(self):
        if self.next_proxy is None:
            raise StopIteration()
        current_proxy = self.next_proxy
        if self.repository is not None:
            current_proxy.load(self.repository)
        try:
            self.next_proxy = current_proxy.next.values[0]
        except IndexError:
            # we have reached the last item of the aggregation
            # set next to None so the iterator will stop on the next iteration
            self.next_proxy = None
        return current_proxy



import csv
import os
import re
from collections import defaultdict
from contextlib import contextmanager
from itertools import zip_longest
from pathlib import Path
from typing import Type, NamedTuple, Mapping, Iterable, TextIO, TypeVar

from rdflib import URIRef, Literal
from urlobject import URLObject

from plastron.models import ContentModeledResource
from plastron.models.fedora import FedoraResource
from plastron.models.pcdm import PCDMFile
from plastron.namespaces import umdaccess
from plastron.rdfmapping.descriptors import ObjectProperty, DataProperty
from plastron.rdfmapping.embed import EmbeddedObject
from plastron.rdfmapping.properties import RDFObjectProperty, RDFDataProperty
from plastron.rdfmapping.resources import RDFResourceBase, RDFResource


def not_empty(value):
    """Returns true if `value` is not `None` and is not the empty string."""
    return value is not None and value != ''


def split_escaped(string: str, separator: str = '|') -> list[str]:
    """Split a string on the separator, taking into account escaped instances
    of the separator.

    ```pycon
    >>> split_escaped('foo|bar')
    ['foo', 'bar']

    >>> split_escaped('foo\|bar|alpha')
    ['foo|bar', 'alpha']
    ```
    """  # noqa: W605
    if string is None or string == '':
        return []
    # uses a negative look-behind to only split on separator characters
    # that are NOT preceded by an escape character (the backslash)
    pattern = re.compile(r'(?<!\\)' + re.escape(separator))
    values = pattern.split(string)
    # remove the escape character
    return [re.sub(r'\\(.)', r'\1', v) for v in values]


def join_values(values: list[list[str] | str]) -> str:
    """Join either a list of strings, or a list of lists of strings. A list of
    strings will be separated with "|". A list of lists will be separated by ";".

    A list of strings represents a single property of a single subject that has
    multiple values. A list of lists represents a several instances of the same
    property, each belonging to a different subject. This latter case is used for
    embedded objects.

    ```pycon
    >>> join_values(['foo', 'bar'])
    'foo|bar'

    >>> join_values([['foo', 'bar'], ['delta']])
    'foo|bar;delta'
    ```
    """
    if values is None or len(values) == 0:
        return ''
    elif isinstance(values[0], list):
        return ';'.join(join_values(v) for v in values)
    else:
        return '|'.join(str(v) for v in values)


def build_lookup_index(index_string: str) -> dict[str, dict[int, str]]:
    """Build a lookup dictionary for embedded object properties of an item.

    From this `index_string`:

    ```python
    'author[0]=#alpha;author[1]=#beta;subject[0]=#delta'
    ```

    Returns this dictionary:

    ```python
    {
        'author': {
            0: 'alpha',
            1: 'beta',
        },
        'subject': {
            0: 'delta',
        },
    }
    ```
    """
    index = defaultdict(dict)
    if index_string is None or index_string == '':
        return index

    pattern = re.compile(r'(\w+)\[(\d+)]')
    for entry in index_string.split(';'):
        key, fragment = entry.split('=')
        m = pattern.search(key)
        attr = m[1]
        i = int(m[2])
        index[attr][i] = fragment.lstrip('#')
    return index


def flatten_headers(header_map: dict[str, str | dict], prefix: str = '') -> dict[str, str]:
    """Transform a possibly nested mapping of attribute name to header label into a
    flat mapping of header label to attribute name. Nesting of the attribute names is
    indicated by a ".".

    ```pycon
    >>> header_mapping = {
    ...     'title': 'Title',
    ...     'creator': {
    ...         'label': 'Creator',
    ...         'same_as': 'Creator URI',
    ...     },
    ... }
    >>> flatten_headers(header_mapping)
    {'Title': 'title', 'Creator': 'creator.label', 'Creator URI': 'creator.same_as'}
    ```
    """
    headers = {}
    for attr, header in header_map.items():
        if isinstance(header, dict):
            headers.update(flatten_headers(header, prefix + attr + '.'))
        else:
            headers[header] = prefix + attr
    return headers


class ColumnHeader(NamedTuple):
    """A column header with an optional language."""
    label: str
    """Column header"""
    language: str = None
    """Language (may be `None`)"""

    @classmethod
    def from_string(cls, header: str) -> 'ColumnHeader':
        """Parse the given string to get the label and language.

        ```pycon
        >>> ColumnHeader.from_string('Title [en]')
        ColumnHeader(label='Title', language='en')

        >>> ColumnHeader.from_string('Author')
        ColumnHeader(label='Author', language=None)
        ```
        """
        m = re.match(r'^([^[]*)(?: \[(.*)])?$', header)
        return cls(label=m[1], language=m[2])

    def __str__(self):
        return f'{self.label} [{self.language}]' if self.language is not None else self.label


class ColumnsDict(dict):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.index = []


def flatten(
    description: RDFResourceBase,
    header_map: dict[str, str | dict],
) -> ColumnsDict:
    """Convert an RDF description to a dictionary with `ColumnHeader` keys and list values, and a
    lookup index list for embedded objects. RDF attributes of the description object are mapped to
    the header keys by the `header_map`."""
    columns = ColumnsDict()
    for attr, header in header_map.items():
        if isinstance(header, dict):
            # treat this as an embedded ObjectProperty
            base_prop = getattr(description, attr)
            assert isinstance(base_prop, RDFObjectProperty), f'"{attr}" must be an object property'
            for n, obj in enumerate(base_prop.objects):
                # add this embedded object to the index for this row
                columns.index.append([f'{attr}[{n}]=#{URLObject(obj.uri).fragment}'])
                embedded_columns = flatten(obj, header)
                columns.update(embedded_columns)
        else:
            if description is not None:
                prop = getattr(description, attr)
                if isinstance(prop, RDFDataProperty):
                    # handle language tags on literals
                    for language in set(v.language for v in prop.values):
                        columns[ColumnHeader(label=header, language=language)] = [
                            v for v in prop.values if v.language == language
                        ]
                else:
                    columns[ColumnHeader(label=header)] = list(prop.values)
            else:
                columns[ColumnHeader(label=header)] = []
    return columns


def get_column_headers(headers: Iterable[str], base_header: str) -> list[ColumnHeader]:
    """From `headers`, return a column `ColumnHeader` for each header whose label is `base_header`.

    ```pycon
    >>> get_column_headers(['Title [en]', 'Title [de]', 'Author'], 'Title')
    [ColumnHeader(label='Title', language='en'), ColumnHeader(label='Title', language='de')]

    >>> get_column_headers(['Title [en]', 'Title [de]', 'Author'], 'Author')
    [ColumnHeader(label='Author', language=None)]

    >>> get_column_headers(['Title [en]', 'Title [de]', 'Author'], 'Date')
    []
    ```
    """
    return [
        ColumnHeader.from_string(h)
        for h in headers
        if h == base_header or re.match(base_header + r' \[.*]$', h)
    ]


def get_embedded_params(row: Mapping[str, str], header_labels: Iterable[str]) -> list[dict[str, str]]:
    """From a data row and a set of header labels, construct a list of parameter
    dictionaries for one or more parallel embedded objects.

    For example, for this CSV:

    ```csv
    Title,Subject,Subject URI
    Foo,Linguistics;Philosophy,http://example.com/term/ling;http://example.com/term/phil
    ```

    Supplying the first row to this function, with `header_labels=["Subject", "Subject URI"]`,
    will yield this list:

    ```python
    [
        {
            "Subject": "Linguistics",
            "Subject URI": "http://example.com/ling",
        },
        {
            "Subject": "Philosophy",
            "Subject URI": "http://example.com/phil",
        },
    ]
    ```

    Which is suitable for passing to `unflatten()` for building the embedded objects themselves.
    """
    params = defaultdict(list)
    for header in header_labels:
        for column_header in get_column_headers(row.keys(), header):
            params[str(column_header)].extend(row[str(column_header)].split(';'))

    sub_rows = list(zip_longest(*params.values()))
    return [dict(zip(params.keys(), x)) for x in sub_rows]


def unflatten(
        row_data: Mapping[str, str],
        resource_class: Type[RDFResourceBase],
        header_map: Mapping[str, str | dict],
        index: Mapping[str, Mapping[int, str]] = None,
) -> dict[str, list[Literal | URIRef | EmbeddedObject]]:
    """Transform a mapping of column headers to values (such as would be returned by a
    `csv.DictReader`) into a dictionary of parameters that can be passed to the constructor
    of an RDF description class to create an RDF description object."""
    if index is None:
        index = {}
    params = defaultdict(list)
    for attr, header in header_map.items():
        descriptor = getattr(resource_class, attr)
        if isinstance(header, dict):
            for n, sub_row in enumerate(get_embedded_params(row_data, header_labels=header.values())):
                embedded_params = unflatten(sub_row, descriptor.object_class, header, index)
                if any(embedded_params.values()):
                    params[attr].append(EmbeddedObject(
                        cls=descriptor.object_class,
                        fragment_id=index.get(attr, {}).get(n, None),
                        **embedded_params,
                    ))
        else:
            for column_header in get_column_headers(row_data.keys(), header):
                values = filter(not_empty, split_escaped(row_data.get(str(column_header)), separator='|'))
                if isinstance(descriptor, ObjectProperty):
                    params[attr].extend(URIRef(v) for v in values)
                else:
                    params[attr].extend(get_literal(column_header, descriptor, v) for v in values)
    return params


def get_literal(column_header: ColumnHeader, descriptor: DataProperty, input_value: str) -> Literal:
    """Given a `ColumnHeader`, a data property descriptor, and a string input value,
    return an RDF `Literal` with the appropriate value, language, and datatype.

    The language code can either be given in the `language` attribute of the `ColumnHeader`,
    or in the `input_value` itself. If it is embedded in the `input_value`, it should appear
    at the start of the string prefixed with `@` and surrounded by square brackets:

    ```
    "[@en]dog"
    "[@de]der Hund"
    "[@ja]イヌ"
    ```
    """
    m = re.match(r'^\[@(\w+)]', input_value)
    if m:
        language = m[1]
        value = input_value[len(language) + 3:]
    else:
        language = column_header.language
        value = input_value
    if descriptor.datatype is not None:
        if language is not None:
            raise RuntimeError('Cannot apply a language tag to a column with a defined datatype')
        return Literal(value, datatype=descriptor.datatype)
    else:
        return Literal(value, lang=language)


@contextmanager
def ensure_text_mode(file):
    if 'b' in file.mode:
        # re-open in text mode
        fh = open(file.fileno(), mode=file.mode.replace('b', ''), closefd=False)
        yield fh
        fh.close()
    else:
        # file is already in text mode
        yield file


@contextmanager
def ensure_binary_mode(file):
    if 'b' not in file.mode:
        # re-open in binary mode
        fh = open(file.fileno(), mode=file.mode + 'b', closefd=False)
        yield fh
        fh.close()
    else:
        # file is already in binary mode
        yield file


class Sheet:
    def __init__(self, model: Type[ContentModeledResource]):
        self.model_class = model
        self.headers = list(flatten_headers(self.header_map).keys()) + CSVSerializer.SYSTEM_HEADERS
        self.extra_headers = defaultdict(set)
        self.rows = []

    @property
    def header_map(self):
        return self.model_class.HEADER_MAP

    def write_csv_file(self, file: TextIO):
        # sort and add the new headers that have language names or datatypes
        for header, new_headers in self.extra_headers.items():
            header_index = self.headers.index(header)
            for i, new_header in enumerate(sorted(new_headers), start=1):
                self.headers.insert(header_index + i, new_header)

        # strip out headers that aren't used in any row
        for header in self.headers:
            has_column_values = any([True for row in self.rows if header in row])
            if not has_column_values:
                self.headers.remove(header)

        # write the CSV file;
        # file must be opened in text mode, otherwise csv complains
        # about wanting str and not bytes
        with ensure_text_mode(file) as csv_file:
            csv_writer = csv.DictWriter(csv_file, self.headers, extrasaction='ignore')
            csv_writer.writeheader()
            for row in self.rows:
                csv_writer.writerow(row)


T = TypeVar('T', ContentModeledResource, RDFResource)


class CSVSerializer:
    """Serializer that encodes metadata records with a defined content model as CSV files."""

    SYSTEM_HEADERS = [
        'URI', 'PUBLIC URI', 'CREATED', 'MODIFIED', 'INDEX', 'FILES', 'ITEM_FILES', 'PUBLISH', 'HIDDEN'
    ]

    def __init__(self, directory: str | Path = None):
        self.directory = Path(directory) if directory is not None else Path.cwd()
        """Destination directory for the CSV file(s)"""

        self.sheets = {}
        """Internal accumulator of row data"""

        self.content_type = 'text/csv'
        self.file_extension = '.csv'

    def __enter__(self):
        self.rows = []
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.finish()

    def write(
            self,
            resource: T,
            files: Iterable = None,
            binaries_dir: str = '',
            public_url: str = None,
    ) -> dict[str, str]:
        """
        Serializes the given resource as a CSV row using the `flatten()` function. The resulting row is
        added to an internal accumulator. The CSV file or files themselves are not actually written until
        the `finish()` method is called.
        """
        resource_class = type(resource)

        if resource_class not in self.sheets:
            self.sheets[resource_class] = Sheet(model=resource_class)
        sheet = self.sheets[resource_class]

        columns = flatten(resource, resource_class.HEADER_MAP)
        for header in columns.keys():
            if header.language is not None:
                sheet.extra_headers[header.label].add(str(header))

        row = {str(k): join_values(v) for k, v in columns.items()}
        row['URI'] = str(resource.uri)
        row['INDEX'] = join_values(columns.index)
        if files is not None:
            row['FILES'] = ';'.join(
                os.path.join(binaries_dir, file.describe(PCDMFile).filename.value) for file in files)
        fedora_resource = resource.redescribe(FedoraResource)
        row['CREATED'] = str(fedora_resource.created.value)
        row['MODIFIED'] = str(fedora_resource.last_modified.value)
        if public_url is not None:
            row['PUBLIC URI'] = public_url

        # set publication and visibility state
        row['PUBLISH'] = str(umdaccess.Published in resource.rdf_type.values)
        row['HIDDEN'] = str(umdaccess.Hidden in resource.rdf_type.values)

        sheet.rows.append(row)

        return row

    LANGUAGE_NAMES = {
        'ja': 'Japanese',
        'ja-latn': 'Japanese (Romanized)'
    }
    LANGUAGE_CODES = {name: code for code, name in LANGUAGE_NAMES.items()}

    DATATYPE_NAMES = {
        URIRef('http://id.loc.gov/datatypes/edtf'): 'EDTF',
        URIRef('http://www.w3.org/2001/XMLSchema#date'): 'Date'
    }
    DATATYPE_URIS = {name: uri for uri, name in DATATYPE_NAMES.items()}

    def finish(self):
        """
        Writes the actual CSV file(s). For each detected content model, this will write
        a CSV file with the name `{ModelName}_metadata.csv` to the directory `directory`.

        Raises an `EmptyItemListError` if there are no items to export.
        """
        if len(self.sheets) == 0:
            raise EmptyItemListError()

        for resource_class, sheet in self.sheets.items():
            metadata_filename = resource_class.__name__ + '_metadata.csv'
            with (self.directory / metadata_filename).open(mode='w') as metadata_file:
                # write a CSV file for this model
                sheet.write_csv_file(metadata_file)


class EmptyItemListError(Exception):
    pass



"""[hOCR](https://kba.github.io/hocr-spec/1.2/) OCR classes"""

import re
from typing import Iterable

# noinspection PyProtectedMember
from lxml.etree import _ElementTree, _Element

from plastron.ocr.core import XYWH, BBox, Scale, RegionBase, OCRResource, BlockRegion, LineRegion, WordRegion

XMLNS = {'html': 'http://www.w3.org/1999/xhtml'}


class HOCRResource(OCRResource):
    """hOCR OCR HTML resource"""
    def __init__(self, doc: _ElementTree, image_resolution: tuple[int, int]):
        self.doc = doc
        capabilities_element = self.doc.xpath('//html:meta[@name="ocr-capabilities"]', namespaces=XMLNS)[0]
        self.capabilities = capabilities_element.get('content').split(' ')
        self.scale = Scale.from_resolution(image_resolution, 'pixel')

    def get_block_nodes(self) -> Iterable[_Element]:
        """Get the `.ocr_carea` descendant elements of this resource."""
        return self.doc.xpath('//html:*[@class="ocr_carea"]', namespaces=XMLNS)

    def get_block_node(self, identifier: str) -> _Element:
        """Get the `.ocr_carea` descendant element whose `@id` matches
        the given `identifier`."""
        return self.doc.xpath("//html:*[@class='ocr_carea'][@id=$id]", id=identifier, namespaces=XMLNS)[0]

    def get_block(self, node: _Element) -> 'ContentArea':
        """Get the `ContentArea` object wrapping the given `.ocr_carea` element."""
        return ContentArea(node, self.scale)


class HOCRRegion(RegionBase):
    """Region within an `HOCRResource`."""
    def __init__(self, element: _Element, scale: Scale):
        self.element = element
        self.scale = scale
        self.id = self.element.get('id')
        properties = self.element.get('title')
        match = re.match(r'bbox (\d+) (\d+) (\d+) (\d+)', properties)
        self._bbox = BBox(*(int(c) for c in match.groups())).scale(self.scale)
        self._xywh = XYWH.from_bbox(self._bbox)


class ContentArea(HOCRRegion, BlockRegion):
    """Wraps an hOCR `.ocr_carea` element."""

    def get_line_nodes(self) -> Iterable[_Element]:
        """Get the `.ocr_line` child elements of this area."""
        return self.element.xpath('.//html:*[@class="ocr_line"]', namespaces=XMLNS)

    def get_line(self, node: _Element) -> 'Line':
        """Get a `Line` object wrapping the given `.ocr_line` element."""
        return Line(node, self.scale)


class Line(HOCRRegion, LineRegion):
    """Wraps an hOCR `.ocr_line` element."""

    def get_word_nodes(self) -> Iterable[_Element]:
        """Get the `.ocrx_word` child elements of this line."""
        return self.element.xpath('.//html:*[@class="ocrx_word"]', namespaces=XMLNS)

    def get_word(self, node: _Element) -> 'Word':
        """Get a `Word` object wrapping the given `.ocrx_word` element."""
        return Word(node, self.scale)


class Word(HOCRRegion, WordRegion):
    """Wraps an hOCR `.ocrx_word` element."""

    def get_content(self) -> str:
        """Get the string content of this word's element."""
        return self.element.text



import re

from edtf_validate.valid_edtf import is_valid as is_valid_edtf
from iso639.language import Language, LanguageNotFoundError


def is_edtf_formatted(value):
    """an EDTF-formatted date"""
    # Allow blank values
    if str(value).strip() == "":
        return True
    return is_valid_edtf(str(value))


def is_valid_iso639_code(value):
    """a valid ISO-639 language code"""
    try:
        Language.match(value)
        return True
    except LanguageNotFoundError:
        return False


def is_handle(value: str) -> bool:
    """a handle URI"""
    return bool(re.match('hdl:[^/]+/.*', value))


def is_iso_8601_date(value: str) -> bool:
    """an ISO 8601 date string (YYYY-MM-DD)"""
    return bool(re.match(r'^\d\d\d\d-\d\d-\d\d$', value))



from plastron.models.letter import Letter

base_uri = 'http://example.com/xyz'


def test_letter_invalid_with_no_fields():
    letter = Letter()
    assert not letter.is_valid


def test_letter_valid_with_only_required_fields():
    letter = Letter()

    # Only provide required fields
    letter.identifier = 'test_letter'
    letter.object_type = 'http://purl.org/dc/dcmitype/Text'
    letter.rights = 'http://vocab.lib.umd.edu/rightsStatement#InC-EDU'
    letter.title = 'Test Letter'
    letter.description = 'Test Letter Description'
    letter.language = 'en'
    letter.part_of = 'http://fedora.info/definitions/v4/repository#inaccessibleResource'
    letter.bibliographic_citation = 'Test Bibliographic Citation'
    letter.rights_holder = 'Test Rights Holder'
    letter.type = 'http://purl.org/dc/dcmitype/Text'
    letter.extent = '1 page'

    assert letter.is_valid



from plastron.handles import HandleBearingResource
from plastron.models import ContentModeledResource
from plastron.models.authorities import Agent, Subject, Place, DCMI_TYPES, UMD_RIGHTS_STATEMENTS, UMD_FORMATS, \
    UMD_ARCHIVAL_COLLECTIONS, UMD_PRESENTATION_SETS, UMD_TERMS_OF_USE_STATEMENTS
from plastron.models.fedora import FedoraResource
from plastron.models.page import File, Page
from plastron.models.pcdm import PCDMObject
from plastron.namespaces import dc, dcterms, edm, pcdm, ore, schema, umdtype, umd
from plastron.rdfmapping.decorators import rdf_type
from plastron.rdfmapping.descriptors import ObjectProperty, DataProperty
from plastron.rdfmapping.resources import RDFResource
from plastron.validation.rules import is_edtf_formatted, is_valid_iso639_code
from plastron.validation.vocabularies import ControlledVocabularyProperty


@rdf_type(pcdm.Object)
class Stub(RDFResource):
    title = DataProperty(dcterms.title, required=True)
    identifier = DataProperty(dcterms.identifier, required=True)
    member_of = ObjectProperty(pcdm.memberOf)


@rdf_type(pcdm.Collection)
class AdminSet(ContentModeledResource, RDFResource):
    model_name = 'AdminSet'
    is_top_level = False

    title = DataProperty(dcterms.title, required=True)


@rdf_type(umd.Item)
class Item(ContentModeledResource, PCDMObject, HandleBearingResource, FedoraResource):
    model_name = 'Item'
    is_top_level = True

    member_of = ObjectProperty(pcdm.memberOf)
    has_file = ObjectProperty(pcdm.hasFile, repeatable=True, cls=File)
    has_member = ObjectProperty(pcdm.hasMember, repeatable=True, cls=Page)
    object_type = ControlledVocabularyProperty(dcterms.type, required=True, vocab=DCMI_TYPES)
    identifier = DataProperty(dcterms.identifier, required=True, repeatable=True)
    rights = ControlledVocabularyProperty(dcterms.rights, required=True, vocab=UMD_RIGHTS_STATEMENTS)
    title = DataProperty(dcterms.title, required=True)
    format = ControlledVocabularyProperty(edm.hasType, vocab=UMD_FORMATS)
    archival_collection = ControlledVocabularyProperty(dcterms.isPartOf, vocab=UMD_ARCHIVAL_COLLECTIONS)
    presentation_set = ControlledVocabularyProperty(ore.isAggregatedBy, repeatable=True, vocab=UMD_PRESENTATION_SETS)
    date = DataProperty(dc.date, validate=is_edtf_formatted)
    description = DataProperty(dcterms.description)
    alternate_title = DataProperty(dcterms.alternative, repeatable=True)
    creator = ObjectProperty(dcterms.creator, repeatable=True, embed=True, cls=Agent)
    contributor = ObjectProperty(dcterms.contributor, repeatable=True, embed=True, cls=Agent)
    publisher = ObjectProperty(dcterms.publisher, repeatable=True, embed=True, cls=Agent)
    audience = ObjectProperty(dcterms.audience, repeatable=True, embed=True, cls=Agent)
    location = ObjectProperty(dcterms.spatial, repeatable=True, embed=True, cls=Place)
    extent = DataProperty(dcterms.extent, repeatable=True)
    subject = ObjectProperty(dcterms.subject, repeatable=True, embed=True, cls=Subject)
    language = DataProperty(dc.language, repeatable=True, validate=is_valid_iso639_code)
    rights_holder = ObjectProperty(dcterms.rightsHolder, repeatable=True, embed=True, cls=Agent)
    terms_of_use = ControlledVocabularyProperty(dcterms.license, vocab=UMD_TERMS_OF_USE_STATEMENTS)
    copyright_notice = DataProperty(schema.copyrightNotice)
    bibliographic_citation = DataProperty(dcterms.bibliographicCitation)
    accession_number = DataProperty(dcterms.identifier, datatype=umdtype.accessionNumber)

    HEADER_MAP = {
        'object_type': 'Object Type',
        'identifier': 'Identifier',
        'rights': 'Rights Statement',
        'title': 'Title',
        'format': 'Format',
        'archival_collection': 'Archival Collection',
        'presentation_set': 'Presentation Set',
        'date': 'Date',
        'description': 'Description',
        'alternate_title': 'Alternate Title',
        'creator': {
            'label': 'Creator',
            'same_as': 'Creator URI',
        },
        'contributor': {
            'label': 'Contributor',
            'same_as': 'Contributor URI',
        },
        'publisher': {
            'label': 'Publisher',
            'same_as': 'Publisher URI',
        },
        'audience': {
            'label': 'Audience',
            'same_as': 'Audience URI',
        },
        'location': {
            'label': 'Location',
        },
        'extent': 'Extent',
        'subject': {
            'label': 'Subject',
        },
        'language': 'Language',
        'rights_holder': {
            'label': 'Rights Holder',
        },
        'terms_of_use': 'Terms of Use',
        'copyright_notice': 'Copyright Notice',
        'bibliographic_citation': 'Collection Information',
        'accession_number': 'Accession Number',
        'handle': 'Handle',
    }

    def __str__(self):
        return str(self.title)
